# Copyright (C) 2023,2024,2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from subsonic_connector.album import Album
from subsonic_connector.album_list import AlbumList
from subsonic_connector.genre import Genre
from subsonic_connector.song import Song
from subsonic_connector.artist import Artist
from subsonic_connector.playlist import Playlist
from subsonic_connector.response import Response
from subsonic_connector.list_type import ListType
from subsonic_connector.playlist_entry import PlaylistEntry

from item_identifier import ItemIdentifier
from item_identifier_key import ItemIdentifierKey
from element_type import ElementType

from album_util import MultiCodecAlbum
from album_util import AlbumTracks
from album_util import strip_codec_from_album
from album_util import get_last_path_element
from album_util import has_year
from album_util import get_album_year_str
from album_util import get_dir_from_path
from album_util import get_album_base_path
import audio_codec

import art_retriever
import config
import connector_provider
import identifier_util
import upnp_util
import subsonic_util
import codec
import constants
import cache_actions
import metadata_converter
from album_metadata import AlbumMetadata
from metadata_model import AlbumMetadataModel

from option_key import OptionKey
from option_util import get_option

import upmplgutils
import upmpdmeta

from msgproc_provider import msgproc

import os
from typing import Optional

import persistence


def artist_entry_for_album(objid, album: Album) -> dict[str, any]:
    msgproc.log(f"artist_entry_for_album creating artist entry for album with album_id: [{album.getId()}]")
    artist_identifier: ItemIdentifier = ItemIdentifier(
        name=ElementType.ARTIST.element_name,
        value=album.getArtistId())
    id: str = identifier_util.create_objid(
        objid=objid,
        id=identifier_util.create_id_from_identifier(artist_identifier))
    artist_entry_title: str = subsonic_util.get_album_display_artist(album=album)
    if album.getArtistId() and config.get_config_param_as_bool(constants.ConfigParam.SHOW_ARTIST_ID):
        msgproc.log(f"artist_entry_for_album: Adding [{album.getArtistId()}] to [{artist_entry_title}]")
        artist_entry_title = f"{artist_entry_title} [{album.getArtistId()}]"
    artist: Artist = subsonic_util.try_get_artist(album.getArtistId())
    if config.get_config_param_as_bool(constants.ConfigParam.SHOW_ARTIST_MB_ID):
        # we want the artist mb id, so we load the artist
        artist_mb_id: str = (subsonic_util.get_artist_musicbrainz_id(artist)
                             if artist
                             else None)
        if artist_mb_id:
            if config.get_config_param_as_bool(constants.ConfigParam.SHOW_ARTIST_MB_ID_AS_PLACEHOLDER):
                artist_entry_title = f"{artist_entry_title} [mb]"
            else:
                artist_entry_title = f"{artist_entry_title} [mb:{artist_mb_id}]"
    # add album count
    album_count: int = artist.getAlbumCount() if artist else 0
    artist_entry_title = f"{artist_entry_title} [{album_count}]"
    artist_entry: dict[str, any] = (upmplgutils.direntry(
        id=id,
        pid=objid,
        title=artist_entry_title))
    if artist_entry:
        art_uri: str = None
        # does the artist has coverArt?
        artist_cover_art: str = subsonic_util.get_artist_cover_art(artist)
        # if not already found from the artist, we try to get an album cover for the artist entry
        if artist_cover_art:
            art_uri = subsonic_util.build_cover_art_url(item_id=artist_cover_art, force_save=True)
        else:
            art_uri = art_retriever.get_album_cover_art_url_by_artist_id(artist_id=album.getArtistId())
        upnp_util.set_album_art_from_uri(album_art_uri=art_uri, target=artist_entry)
        cache_actions.on_album(album=album)
    return artist_entry


def genre_artist_to_entry(
        objid,
        genre: str,
        artist_id: str,
        artist_name: str,
        album_cover_art: str = None) -> dict[str, any]:
    msgproc.log(f"genre_artist_to_entry genre:[{genre}] artist_id:[{artist_id}] artist_name:[{artist_name}]")
    identifier: ItemIdentifier = ItemIdentifier(
        ElementType.GENRE_ARTIST.element_name,
        artist_id)
    identifier.set(ItemIdentifierKey.GENRE_NAME, genre)
    id: str = identifier_util.create_objid(
        objid=objid,
        id=identifier_util.create_id_from_identifier(identifier))
    entry = upmplgutils.direntry(
        id,
        objid,
        artist_name)
    if artist_id:
        entry: dict[str, any] = artist_to_entry_raw(
            objid=objid,
            artist_id=artist_id,
            artist_entry_name=artist_name,
            album_cover_art=album_cover_art)
    return entry


def album_to_navigable_entry(
        objid,
        album: Album,
        album_metadata: AlbumMetadata = None,
        options: dict[str, any] = {}) -> dict[str, any]:
    title: str = album.getTitle()
    album_mbid: str = subsonic_util.get_album_musicbrainz_id(album=album)
    if album_metadata is None:
        album_metadata = persistence.get_album_metadata(album_id=album.getId())
    # still None? save it!
    if album_metadata is None:
        album_metadata, _ = persistence.save_album_metadata(
            album_metadata=metadata_converter.build_album_metadata(album=album),
            context="album_to_navigable_entry",
            force_insert=True)
    album_quality_badge: str = album_metadata.quality_badge if album_metadata else None
    album_lossless_status: str = album_metadata.get_value(AlbumMetadataModel.ALBUM_LOSSLESS_STATUS)
    album_avg_bitrate: int = album_metadata.get_value(AlbumMetadataModel.ALBUM_AVERAGE_BITRATE)
    if album_quality_badge is not None and album_lossless_status is not None and album_avg_bitrate is not None:
        # see if we need to add bitrate
        if not audio_codec.get_lossless_status_by_value(v=album_lossless_status) == audio_codec.LosslessStatus.LOSSLESS:
            # we add bitrate to badge
            album_quality_badge = f"{album_quality_badge} ({album_avg_bitrate}kb/s)"
    album_version: str = subsonic_util.get_album_version(album)
    # explicit?
    title = subsonic_util.append_explicit_if_needed(title, album)
    album_date_for_sorting: str = subsonic_util.get_album_date_for_sorting(album)
    if config.get_config_param_as_bool(constants.ConfigParam.DUMP_ALBUM_SORTABLE_DATE):
        msgproc.log(f"Album [{album.getId()}] [{album.getTitle()}] "
                    f"by [{subsonic_util.get_album_display_artist(album=album)}] "
                    f"Sortable Date [{album_date_for_sorting}]")
    prepend_number: int = get_option(options=options, option_key=OptionKey.PREPEND_ENTRY_NUMBER_IN_ALBUM_TITLE)
    if prepend_number:
        title = f"[{prepend_number:02}] {title}"
    artist: str = subsonic_util.get_album_display_artist(album=album)
    identifier: ItemIdentifier = ItemIdentifier(ElementType.NAVIGABLE_ALBUM.element_name, album.getId())
    # ask to skip artist?
    skip_artist_id: str = get_option(options=options, option_key=OptionKey.SKIP_ARTIST_ID)
    if skip_artist_id:
        identifier.set(ItemIdentifierKey.SKIP_ARTIST_ID, skip_artist_id)
    id: str = identifier_util.create_objid(
        objid=objid,
        id=identifier_util.create_id_from_identifier(identifier))
    if has_year(album):
        if config.get_config_param_as_bool(constants.ConfigParam.APPEND_YEAR_TO_ALBUM_CONTAINER):
            title = f"{title} [{get_album_year_str(album)}]"
    else:
        msgproc.log(f"Cannot find year for album [{album.getId()}] "
                    f"[{album.getTitle()}] by "
                    f"[{subsonic_util.get_album_display_artist(album=album)}]")
    # append genre if allowed
    title = subsonic_util.append_genre_to_artist_entry_name_if_allowed(
        entry_name=title,
        album=album,
        config_getter=(lambda: config.get_config_param_as_bool(constants.ConfigParam.ALLOW_GENRE_IN_ALBUM_CONTAINER)))
    title = subsonic_util.append_album_badge_to_album_title(
        current_albumtitle=title,
        album_quality_badge=album_quality_badge,
        album_entry_type=constants.AlbumEntryType.ALBUM_CONTAINER,
        is_search_result=False)
    title = subsonic_util.append_album_version_to_album_title(
        current_albumtitle=title,
        album_version=album_version,
        album_entry_type=constants.AlbumEntryType.ALBUM_CONTAINER,
        is_search_result=False)
    append_artist: bool = (config.get_config_param_as_bool(
                           constants.ConfigParam.ALLOW_APPEND_ARTIST_IN_ALBUM_CONTAINER) and
                           get_option(options=options, option_key=OptionKey.APPEND_ARTIST_IN_ALBUM_TITLE))
    if append_artist:
        artist: str = subsonic_util.get_album_display_artist(album=album)
        if artist:
            title = f"{title} - {artist}"
    entry_title: str = title
    entry_title = subsonic_util.append_album_id_to_album_title(
        current_albumtitle=entry_title,
        album_id=album.getId(),
        album_entry_type=constants.AlbumEntryType.ALBUM_CONTAINER,
        is_search_result=False)
    if config.get_config_param_as_bool(constants.ConfigParam.SHOW_ALBUM_MBID_IN_ALBUM_CONTAINER):
        # available here?
        album_mbid: str = subsonic_util.get_album_musicbrainz_id(album)
        if album_mbid:
            if config.get_config_param_as_bool(constants.ConfigParam.DUMP_ACTION_ON_MB_ALBUM_CACHE):
                msgproc.log(f"Album album_mbid for [{album.getId()}] -> [{album_mbid}]")
            # display it if requested
            if config.get_config_param_as_bool(constants.ConfigParam.SHOW_ALBUM_MB_ID_AS_PLACEHOLDER):
                entry_title = f"{entry_title} [mb]"
            else:
                entry_title = f"{entry_title} [mb:{album_mbid}]"
    # anomalies
    show_album_genre_information(album)
    entry: dict[str, any] = upmplgutils.direntry(
        id=id,
        pid=objid,
        title=entry_title,
        artist=artist)
    upnp_util.set_upmpd_meta(upmpdmeta.UpMpdMeta.ALBUM_QUALITY, album_quality_badge, entry)
    subsonic_util.set_album_metadata(
        album=album,
        target=entry,
        album_metadata=album_metadata)
    upnp_util.set_album_art_from_uri(
        album_art_uri=subsonic_util.build_cover_art_url(item_id=album.getCoverArt()),
        target=entry)
    upnp_util.set_album_id(album.getId(), entry)
    if config.get_config_param_as_bool(constants.ConfigParam.SET_CLASS_TO_ALBUM_FOR_NAVIGABLE_ALBUM):
        upnp_util.set_class_album(entry)
    else:
        upnp_util.set_class(upnp_class="object.container", target=entry)
    # cache_actions.on_album(album=album)
    return entry


def show_album_genre_information(album: Album):
    if not config.get_config_param_as_bool(constants.ConfigParam.DUMP_ALBUM_GENRE):
        return
    genre_list: list[str] = album.getGenres()
    if not genre_list or len(genre_list) == 0:
        msgproc.log(f"WARN: Album [{album.getId()}] [{album.getTitle()}] "
                    f"by [{subsonic_util.get_album_display_artist(album=album)}] has no genres")
    else:
        msgproc.log(f"Album [{album.getId()}] [{album.getTitle()}] by "
                    f"[{subsonic_util.get_album_display_artist(album=album)}] "
                    f"has genres [{genre_list}]")


def genre_to_entry(
        objid,
        current_genre: Genre) -> dict[str, any]:
    verbose: bool = config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING)
    genre_name: str = current_genre.getName()
    genre_art: str = None
    genre_art_url: str = None
    # first, query the db
    album_metadata: AlbumMetadata = persistence.get_random_album_by_genre(
        genre_name=genre_name)
    if album_metadata:
        if verbose:
            msgproc.log("genre_to_entry got from db "
                        f"album_id [{album_metadata.album_id}] "
                        f"cover_art [{album_metadata.album_cover_art}]")
        genre_art = album_metadata.album_cover_art
        genre_art_url = (subsonic_util.build_cover_art_url(item_id=genre_art)
                         if genre_art
                         else None)
    if (not genre_art and
            config.get_config_param_as_bool(constants.ConfigParam.GENRE_VIEW_SEARCH_ALBUMS_FOR_COVER_ART)):
        # load up to 5 albums
        res: Response[AlbumList] = connector_provider.get().getAlbumList(
            ltype=ListType.BY_GENRE,
            genre=genre_name,
            size=5)
        if not res or not res.isOk():
            msgproc.log(f"Cannot get albums by genre [{genre_name}]")
        album_list: AlbumList = res.getObj()
        msgproc.log(f"Loaded [{len(album_list.getAlbums()) if album_list and album_list.getAlbums() else 0}] "
                    f"albums for genre [{genre_name}]")
        if album_list and len(album_list.getAlbums()) > 0:
            # look for one with a cover art
            album: Album
            for album in album_list.getAlbums():
                if not album.getCoverArt():
                    continue
                genre_art = album.getCoverArt()
                genre_art_url = subsonic_util.build_cover_art_url(item_id=genre_art)
    identifier: ItemIdentifier = ItemIdentifier(
        ElementType.GENRE.element_name,
        current_genre.getName())
    id: str = identifier_util.create_objid(
        objid,
        identifier_util.create_id_from_identifier(identifier))
    entry = upmplgutils.direntry(id, objid, genre_name)
    upnp_util.set_album_art_from_uri(genre_art_url, entry)
    return entry


def maybe_append_roles(entry_name: str, artist: Artist) -> str:
    if config.get_config_param_as_bool(constants.ConfigParam.APPEND_ROLES_TO_ARTIST):
        artist_roles: list[str] = subsonic_util.get_artist_roles(artist=artist)
        disp: list[str] = list(map(subsonic_util.name_key_to_display, artist_roles))
        return f"{entry_name} [{', '.join(disp)}]"
    else:
        return entry_name


def artist_to_entry(
        objid,
        artist: Artist,
        entry_name: str = None,
        cover_art: str = None,
        additional_identifier_properties: dict[ItemIdentifierKey, any] = {},
        options: dict[str, any] = {}) -> dict[str, any]:
    verbose: bool = config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING)
    cover_art: str = cover_art if cover_art else subsonic_util.get_artist_cover_art(artist)
    artist_roles: list[str] = subsonic_util.get_artist_roles(artist=artist)
    if verbose:
        msgproc.log(f"artist_to_entry artist [{artist.getId()}] [{artist.getName()}] -> "
                    f"roles [{artist_roles}] "
                    f"coverArt [{cover_art}]")
    select_artist_entry_name: str = entry_name if entry_name else maybe_append_roles(artist.getName(), artist)
    artist_entry: dict[str, any] = artist_to_entry_raw(
        objid=objid,
        artist_id=artist.getId(),
        artist_entry_name=select_artist_entry_name,
        artist_cover_art=cover_art,
        additional_identifier_properties=additional_identifier_properties,
        options=options)
    subsonic_util.set_artist_metadata(artist=artist, target=artist_entry)
    return artist_entry


def artist_to_entry_raw(
        objid,
        artist_id: str,
        artist_entry_name: str,
        artist_cover_art: str = None,
        album_cover_art: str = None,
        additional_identifier_properties: dict[ItemIdentifierKey, any] = {},
        options: dict[str, any] = {}) -> dict[str, any]:
    verbose: bool = config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING)
    identifier: ItemIdentifier = ItemIdentifier(
        ElementType.ARTIST.element_name,
        artist_id)
    k: ItemIdentifierKey
    for k, v in additional_identifier_properties.items() if additional_identifier_properties else {}.items():
        msgproc.log(f"Adding [{k}]: [{v}]")
        identifier.set(k, v)
    id: str = identifier_util.create_objid(
        objid=objid,
        id=identifier_util.create_id_from_identifier(identifier))
    entry = upmplgutils.direntry(id=id, pid=objid, title=artist_entry_name)
    skip_art: bool = get_option(options=options, option_key=OptionKey.SKIP_ART)
    album_art_uri: str = (subsonic_util.build_cover_art_url(item_id=artist_cover_art)
                          if artist_cover_art is not None
                          else None)
    if verbose:
        msgproc.log(f"artist_to_entry_raw for artist_id [{artist_id}] [{artist_entry_name}] -> "
                    f"artist_cover_art [{artist_cover_art}]")
    if not album_art_uri and (not skip_art and artist_id):
        # first we try in the cache ...
        if verbose:
            msgproc.log(f"artist_to_entry_raw retrieving cover art for artist_id [{artist_id}] in metadata cache ...")
        artist_metadata: persistence.ArtistMetadata = persistence.get_artist_metadata(artist_id=artist_id)
        if artist_metadata and artist_metadata.artist_cover_art:
            if verbose:
                msgproc.log(f"artist_to_entry_raw artist_id [{artist_id}] "
                            f"metadata item found -> cover_art [{artist_metadata.artist_cover_art}]")
            album_art_uri = subsonic_util.build_cover_art_url(item_id=artist_metadata.artist_cover_art)
            if verbose:
                msgproc.log(f"artist_to_entry_raw artist_id [{artist_id}] -> album_art_uri [{album_art_uri}]")
            if verbose:
                msgproc.log(f"artist_to_entry_raw found cached cover_art for artist_id [{artist_id}] -> "
                            f"[{artist_metadata.artist_cover_art}] -> "
                            f"album_art_uri [{album_art_uri}]")
        else:
            # was the function argument album_cover_art provided?
            if album_cover_art:
                if verbose:
                    msgproc.log(f"artist_to_entry_raw for artist_id [{artist_id}] "
                                f"we are using album cover art [{album_cover_art}] ...")
                    album_art_uri = subsonic_util.build_cover_art_url(item_id=album_cover_art)
            else:
                # find art from albums
                if verbose:
                    msgproc.log("artist_to_entry_raw retrieving cover art for "
                                f"artist_id [{artist_id}] from albums, this is slow ...")
                album_art_uri = art_retriever.get_album_cover_art_url_by_artist_id(artist_id)
    if album_art_uri:
        upnp_util.set_album_art_from_uri(
            album_art_uri=album_art_uri,
            target=entry)
    upnp_util.set_class_artist(entry)
    subsonic_util.set_artist_metadata_by_artist_id(
        artist_id=artist_id,
        target=entry)
    return entry


def artist_initial_to_entry(
        objid,
        artist_initial: str,
        options: dict[str, any] = dict()) -> dict[str, any]:
    encoded_artist_initial: str = codec.base64_encode(artist_initial)
    identifier: ItemIdentifier = ItemIdentifier(
        ElementType.ARTIST_BY_INITIAL.element_name,
        encoded_artist_initial)
    id: str = identifier_util.create_objid(
        objid=objid,
        id=identifier_util.create_id_from_identifier(identifier))
    entry: dict[str, any] = upmplgutils.direntry(id, objid, artist_initial)
    return entry


def build_intermediate_url(track_id: str, suffix: str) -> str:
    verbose: bool = config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING)
    enable_track_intermediate_url: bool = config.get_config_param_as_bool(constants.ConfigParam.ENABLE_TRACK_INTERMEDIATE_URL)
    if verbose:
        msgproc.log(f"build_intermediate_url enable_track_intermediate_url [{enable_track_intermediate_url}] "
                    f"track_id [{track_id}] "
                    f"suffix [{suffix}] "
                    f"transcode_format [{config.get_transcode_codec()}] "
                    f"transcode max bitrate [{config.get_transcode_max_bitrate()}]")
    if enable_track_intermediate_url:
        http_host_port = os.environ["UPMPD_HTTPHOSTPORT"]
        url = (f"http://{http_host_port}/{constants.PluginConstant.PLUGIN_NAME.value}"
               f"/track/version/1/trackId/{track_id}")
        if config.get_config_param_as_bool(constants.ConfigParam.LOG_INTERMEDIATE_URL):
            msgproc.log(f"intermediate_url for track_id {track_id} -> [{url}]")
        return url
    else:
        tr_format: str = config.get_transcode_codec()
        tr_bitrate: int = config.get_transcode_max_bitrate()
        if tr_format and suffix and tr_format.lower() == suffix.lower():
            # same suffix, avoid transcoding
            if verbose:
                msgproc.log(f"build_intermediate_url transcoding skipped because "
                            f"transcoding suffix is [{suffix}] "
                            f"and transcoding format is [{tr_format}]")
            tr_format = None
            tr_bitrate = None
        return connector_provider.get().buildSongUrl(
            song_id=track_id,
            format=tr_format,
            max_bitrate=tr_bitrate)


def set_song_quality_flags(song: Song | PlaylistEntry, entry: dict[str, any]):
    verbose: bool = config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING)
    # declaring channel count, bit depth, sample rate and bit rate
    cc: int = 2
    bd: int = 0
    sr: int = 0
    br: int = 0
    mimetype: str = ""
    cc = subsonic_util.get_song_channel_count(song=song)
    upnp_util.set_channel_count(cc, entry)
    bd = subsonic_util.get_song_bit_depth(song=song)
    sr = subsonic_util.get_song_sampling_rate(song=song)
    upnp_util.set_sample_rate(sr, entry)
    # will transcoding be applied?
    tr_format: str = config.get_transcode_codec()
    tr_bitrate: int = config.get_transcode_max_bitrate()
    transcoding_applies: bool = tr_format is not None and song.getSuffix() and tr_format.lower() != song.getSuffix().lower()
    if verbose:
        msgproc.log(f"set_song_quality_flags transcoding applies [{transcoding_applies}] "
                    f"song_suffix [{song.getSuffix()}] "
                    f"tr_format [{tr_format}] -> "
                    f"[{transcoding_applies}]")
    if transcoding_applies:
        # transcoding applies
        # we set values from transcoding configuration
        br = tr_bitrate
        upnp_util.set_bit_rate(tr_bitrate, entry)
        # set bitdepth
        tr_bitdepth: int = constants.get_default_bitdepth_by_codec(tr_format)
        if verbose:
            msgproc.log(f"set_song_quality_flags tr_bitdepth for [{tr_format}] -> [{tr_bitdepth}]")
        # update bd with tr_bitdepth if meaningful
        bd = tr_bitdepth if tr_bitdepth else bd
        upnp_util.set_bit_depth(bd, entry)
        # guess mime type from suffx
        guessed_mimetype: Optional[str] = subsonic_util.get_mime_type_from_extension(tr_format)
        if verbose:
            msgproc.log(f"set_song_quality_flags guessed_mimetype from tr_format [{tr_format}]: [{guessed_mimetype}]")
        mimetype = guessed_mimetype
        upnp_util.set_mimetype(mimetype, entry)
    else:
        # transcoding does not apply
        upnp_util.set_bit_depth(bd, entry)
        br = song.getBitRate()
        upnp_util.set_bit_rate(br, entry)
        # mime type from song itself
        mimetype = song.getContentType()
        upnp_util.set_mimetype(mimetype, entry)
        # size is known
        upnp_util.set_size(subsonic_util.get_song_size(song=song), entry)


def song_to_entry(
        objid,
        song: Song,
        force_cover_art_save: bool = False,
        options: dict[str, any] = {}) -> dict:
    verbose: bool = config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING)
    entry = {}
    identifier: ItemIdentifier = ItemIdentifier(ElementType.SONG.element_name, song.getId())
    id: str = identifier_util.create_objid(
        objid=objid,
        id=identifier_util.create_id_from_identifier(identifier))
    entry['id'] = id
    entry['pid'] = song.getId()
    upnp_util.set_class_music_track(entry)
    song_uri: str = build_intermediate_url(track_id=song.getId(), suffix=song.getSuffix())
    entry['uri'] = song_uri
    title: str = song.getTitle()
    multi_codec_album: MultiCodecAlbum = get_option(options=options, option_key=OptionKey.MULTI_CODEC_ALBUM)
    if (MultiCodecAlbum.YES == multi_codec_album and
        config.get_config_param_as_bool(constants.ConfigParam.ALLOW_BLACKLIST_CODEC_IN_SONG) and
            (not song.getSuffix().lower() in config.get_whitelist_codecs())):
        title = "{} [{}]".format(title, song.getSuffix())
    upnp_util.set_track_title(title, entry)
    entry['tp'] = 'it'
    entry['discnumber'] = song.getDiscNumber()
    track_num: str = song.getTrack()
    force_track_number: int = get_option(options=options, option_key=OptionKey.FORCE_TRACK_NUMBER)
    if force_track_number:
        track_num = str(force_track_number)
    upnp_util.set_track_number(track_num, entry)
    upnp_util.set_artist(subsonic_util.get_song_display_artist(song=song), entry)
    song_album_artist: str = subsonic_util.get_song_display_album_artist(song=song)
    if song_album_artist:
        if config.get_config_param_as_bool(constants.ConfigParam.ALLOW_SONG_DIDL_ALBUMARTIST):
            if verbose:
                msgproc.log(f"Setting didlfrag with [{song_album_artist}] ...")
            upnp_util.set_didlfrag(
                didlfrag=upnp_util.build_didlfrag(
                    key="upnp:artist",
                    role="AlbumArtist",
                    value=song_album_artist),
                target=entry)
    entry['upnp:album'] = song.getAlbum()
    entry['upnp:genre'] = song.getGenre()
    album_art_uri: str = subsonic_util.build_cover_art_url(item_id=song.getCoverArt(), force_save=force_cover_art_save)
    upnp_util.set_album_art_from_uri(album_art_uri=album_art_uri, target=entry)
    entry['duration'] = str(song.getDuration())
    set_song_quality_flags(song=song, entry=entry)
    if verbose:
        msgproc.log(f"song_to_entry song id [{song.getId()}] -> [{entry}]")
    subsonic_util.set_song_metadata(
        song=song,
        target=entry)
    return entry


def playlist_to_entry(
        objid,
        playlist: Playlist) -> dict[str, any]:
    identifier: ItemIdentifier = ItemIdentifier(
        ElementType.PLAYLIST.element_name,
        playlist.getId())
    id: str = identifier_util.create_objid(
        objid=objid,
        id=identifier_util.create_id_from_identifier(identifier))
    entry = upmplgutils.direntry(id, objid, playlist.getName())
    art_uri: str = (subsonic_util.build_cover_art_url(item_id=playlist.getCoverArt())
                    if playlist.getCoverArt() else None)
    upnp_util.set_album_art_from_uri(album_art_uri=art_uri, target=entry)
    upnp_util.set_class_playlist_container(target=entry)
    return entry


def album_to_entry(
        objid,
        album: Album,
        album_metadata: AlbumMetadata = None,
        options: dict[str, any] = {}) -> dict[str, any]:
    verbose: bool = config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING)
    if album_metadata is None:
        album_metadata = persistence.get_album_metadata(album_id=album.getId())
    # still None? save it!
    if album_metadata is None:
        album_metadata, _ = persistence.save_album_metadata(
            album_metadata=metadata_converter.build_album_metadata(album=album),
            context="album_to_entry",
            force_insert=True)
    is_search_result: bool = get_option(options=options, option_key=OptionKey.SEARCH_RESULT)
    title: str = album.getTitle()
    album_version: str = subsonic_util.get_album_version(album)
    # explicit?
    title = subsonic_util.append_explicit_if_needed(title, album)
    append_artist: bool = (config.get_config_param_as_bool(
                           constants.ConfigParam.ALLOW_APPEND_ARTIST_IN_ALBUM_VIEW) and
                           get_option(options=options, option_key=OptionKey.APPEND_ARTIST_IN_ALBUM_TITLE)
                           if not is_search_result
                           else config.get_config_param_as_bool(constants.ConfigParam.ALLOW_APPEND_ARTIST_IN_SEARCH_RES))
    if append_artist:
        artist: str = subsonic_util.get_album_display_artist(album=album)
        if artist:
            title = f"{artist} - {title}"
    prepend_number: int = get_option(options=options, option_key=OptionKey.PREPEND_ENTRY_NUMBER_IN_ALBUM_TITLE)
    if prepend_number:
        title = f"[{prepend_number:02}] {title}"
    append_year: bool = (config.get_config_param_as_bool(constants.ConfigParam.APPEND_YEAR_TO_ALBUM_SEARCH_RES)
                         if is_search_result
                         else config.get_config_param_as_bool(constants.ConfigParam.APPEND_YEAR_TO_ALBUM_VIEW))
    if append_year and has_year(album):
        title = "{} [{}]".format(title, get_album_year_str(album))
    force_load_quality_info: bool = get_option(
        options=options,
        option_key=OptionKey.FORCE_RELOAD_ALBUM_QUALITY_INFO)
    msgproc.log(f"album_to_entry for [{album.getId()}] force_load_quality_info [{force_load_quality_info}]")
    album_quality_badge: str = album_metadata.quality_badge if album_metadata else None
    any_lossy: bool = album_metadata.get_value(AlbumMetadataModel.ALBUM_LOSSLESS_STATUS) in [
            audio_codec.LosslessStatus.LOSSY,
            audio_codec.LosslessStatus.MIXED]
    if force_load_quality_info:
        # reload quality badge and track quality summary
        album_quality_badge = subsonic_util.calc_song_list_quality_badge(
            song_list=album.getSongs(),
            list_identifier=album.getId())
        album_lossless_status: audio_codec.LosslessStatus = subsonic_util.calc_lossless_status(
            song_info_list=subsonic_util.get_song_info_list(song_list=album.getSongs()))
        any_lossy = album_lossless_status in [audio_codec.LosslessStatus.LOSSY, audio_codec.LosslessStatus.MIXED]
        song_quality_summary: str = subsonic_util.calc_song_quality_summary(song_list=album.getSongs())
        # update metadata
        album_metadata, _ = persistence.save_album_metadata(
            album_metadata=metadata_converter.build_album_metadata(
                album=album,
                quality_badge=album_quality_badge,
                song_quality_summary=song_quality_summary),
            context="album_to_entry",
            force_insert=album_metadata is None)
    if force_load_quality_info and config.get_config_param_as_bool(constants.ConfigParam.APPEND_CODEC_TO_ALBUM):
        msgproc.log(f"album_to_entry for "
                    f"album_id: [{album.getId()}] "
                    f"badge [{album_quality_badge if album_quality_badge else 'not available'}] "
                    "loading songs ...")
        song_list: list[Song] = album.getSongs()
        # load album
        album_tracks: AlbumTracks
        _, album_tracks = subsonic_util.get_album_tracks(album.getId())
        if album_tracks is None:
            return None
        song_list: list[Song] = album_tracks.getSongList()
        codecs: list[str] = []
        whitelist_count: int = 0
        blacklist_count: int = 0
        song: Song
        for song in song_list:
            if not song.getSuffix().lower() in codecs:
                codecs.append(song.getSuffix().lower())
                if not song.getSuffix().lower() in config.get_whitelist_codecs():
                    blacklist_count += 1
                else:
                    whitelist_count += 1
        # show version count if count > 1
        if album_tracks.getAlbumVersionCount() > 1:
            title = "{} [{} versions]".format(title, album_tracks.getAlbumVersionCount())
        # show or not?
        all_whitelisted: bool = len(codecs) == whitelist_count
        if len(codecs) > 1 or not all_whitelisted:
            codecs.sort()
            if len(codecs) == 1:
                title = strip_codec_from_album(title, codecs)
            codecs_str: str = ",".join(codecs)
            # add codecs if more than one or there is no quality_badge
            if (len(codecs) > 1) or (not album_quality_badge or len(album_quality_badge) == 0):
                title = "{} [{}]".format(title, codecs_str)
    # set badge
    if any_lossy:
        # add bitrate to quality badge
        avg_bitrate: int = album_metadata.get_value(AlbumMetadataModel.ALBUM_AVERAGE_BITRATE)
        if not avg_bitrate:
            avg_bitrate = subsonic_util._get_avg_bitrate(song_info_list=subsonic_util.get_song_info_list(album.getSongs()))
        album_quality_badge = f"{album_quality_badge} ({avg_bitrate})"
    title = subsonic_util.append_album_badge_to_album_title(
        current_albumtitle=title,
        album_quality_badge=album_quality_badge,
        album_entry_type=constants.AlbumEntryType.ALBUM_VIEW,
        is_search_result=is_search_result)
    title = subsonic_util.append_album_version_to_album_title(
        current_albumtitle=title,
        album_version=album_version,
        album_entry_type=constants.AlbumEntryType.ALBUM_VIEW,
        is_search_result=is_search_result)
    # show album id
    title = subsonic_util.append_album_id_to_album_title(
        current_albumtitle=title,
        album_id=album.getId(),
        album_entry_type=constants.AlbumEntryType.ALBUM_VIEW,
        is_search_result=is_search_result)
    # musicbrainz?
    album_mbid: str = subsonic_util.get_album_musicbrainz_id(album)
    if verbose:
        msgproc.log(f"Found album_mbid [{album_mbid}] "
                    f"for album [{album.getId()}] "
                    f"[{album.getTitle()}] "
                    f"by [{subsonic_util.get_album_display_artist(album=album)}]")
    show_mbid: bool = config.get_config_param_as_bool(
        constants.ConfigParam.SHOW_ALBUM_MBID_IN_ALBUM_SEARCH_RES
        if is_search_result
        else constants.ConfigParam.SHOW_ALBUM_MBID_IN_ALBUM_VIEW)
    if album_mbid and show_mbid:
        if config.get_config_param_as_bool(constants.ConfigParam.SHOW_ARTIST_MB_ID_AS_PLACEHOLDER):
            title = f"{title} [mb]"
        else:
            title = f"{title} [{album_mbid}]"
    artist = subsonic_util.get_album_display_artist(album=album)
    # cache_actions.on_album(album=album)
    identifier: ItemIdentifier = ItemIdentifier(ElementType.ALBUM.element_name, album.getId())
    id: str = identifier_util.create_objid(
        objid=objid,
        id=identifier_util.create_id_from_identifier(identifier))
    entry: dict[str, any] = upmplgutils.direntry(id, objid, title=title, artist=artist)
    didl_fragment: str = ""
    # add album artist
    if config.get_config_param_as_bool(constants.ConfigParam.SET_ALBUM_ARTIST_ROLE_ALBUMARTIST):
        didl_fragment += upnp_util.build_didlfrag(key="upnp:artist", role="albumartist", value=artist)
    if didl_fragment:
        upnp_util.set_didlfrag(didlfrag=didl_fragment, target=entry)
    # we save the cover art even if it's already there
    cover_art_url: str = subsonic_util.build_cover_art_url(item_id=album.getCoverArt(), force_save=True)
    upnp_util.set_album_art_from_uri(cover_art_url, entry)
    upnp_util.set_album_id(album.getId(), entry)
    upnp_util.set_artist(artist=subsonic_util.get_album_display_artist(album=album), target=entry)
    upnp_util.set_date_from_album(album=album, target=entry)
    upnp_util.set_class_album(entry)
    upnp_util.set_upmpd_meta(upmpdmeta.UpMpdMeta.ALBUM_QUALITY, album_quality_badge, entry)
    subsonic_util.set_album_metadata(
        album=album,
        target=entry,
        album_metadata=album_metadata)
    return entry


def _load_album_version_tracks(
        album: Album,
        album_version_path: str) -> list[Song]:
    track_list: list[Song] = list()
    current_song: Song
    for current_song in album.getSongs():
        song_path: str = get_dir_from_path(current_song.getPath())
        song_path = get_album_base_path(song_path)
        if album_version_path == song_path:
            track_list.append(current_song)
    return track_list


def album_id_to_album_focus(
        objid,
        album: Album) -> dict[str, any]:
    identifier: ItemIdentifier = ItemIdentifier(
        ElementType.ALBUM_FOCUS.element_name,
        album.getId())
    id: str = identifier_util.create_objid(
        objid=objid,
        id=identifier_util.create_id_from_identifier(identifier))
    art_uri = subsonic_util.build_cover_art_url(item_id=album.getCoverArt())
    entry = upmplgutils.direntry(id, objid, "Focus")
    upnp_util.set_album_art_from_uri(album_art_uri=art_uri, target=entry)
    return entry


def artist_id_to_artist_focus(
        objid,
        artist_id: str) -> dict[str, any]:
    identifier: ItemIdentifier = ItemIdentifier(
        ElementType.ARTIST_FOCUS.element_name,
        artist_id)
    id: str = identifier_util.create_objid(
        objid=objid,
        id=identifier_util.create_id_from_identifier(identifier))
    entry = upmplgutils.direntry(id, objid, "Focus")
    return entry


def album_version_to_entry(
        objid,
        current_album: Album,
        version_number: int,
        album_version_path: str,
        codec_set: set[str]) -> dict[str, any]:
    identifier: ItemIdentifier = ItemIdentifier(ElementType.ALBUM_VERSION.element_name, current_album.getId())
    avp_encoded: str = codec.base64_encode(album_version_path)
    identifier.set(ItemIdentifierKey.ALBUM_VERSION_PATH_BASE64, avp_encoded)
    id: str = identifier_util.create_objid(
        objid=objid,
        id=identifier_util.create_id_from_identifier(identifier))
    title: str = f"Version #{version_number}"
    if (config.get_config_param_as_bool(constants.ConfigParam.APPEND_YEAR_TO_ALBUM_VIEW) and
            has_year(current_album)):
        title = "{} [{}]".format(title, get_album_year_str(current_album))
    codecs_str: str = ",".join(codec_set)
    title = "{} [{}]".format(title, codecs_str)
    last_path: str = get_last_path_element(album_version_path)
    title = "{} [{}]".format(title, last_path)
    # album badge on the list of tracks
    track_list: list[Song] = _load_album_version_tracks(
        album=current_album,
        album_version_path=album_version_path)
    album_quality_badge: str = subsonic_util.calc_song_list_quality_badge(song_list=track_list)
    msgproc.log(f"album_version_to_entry album [{current_album.getId()}] -> badge [{album_quality_badge}]")
    if album_quality_badge:
        title = f"{title} [{album_quality_badge}]"
        msgproc.log(f"album_version_to_entry title [{title}]")
    artist = subsonic_util.get_album_display_artist(album=current_album)
    cache_actions.on_album(current_album)
    entry: dict[str, any] = upmplgutils.direntry(id, objid, title=title, artist=artist)
    current_album_cover_art: str = subsonic_util.build_cover_art_url(item_id=current_album.getCoverArt())
    upnp_util.set_album_art_from_uri(current_album_cover_art, entry)
    upnp_util.set_class_album(entry)
    return entry



# Copyright (C) 2023 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64


def encode(name: str) -> str:
    message_bytes: bytes = name.encode('utf-8')
    base64_bytes: bytes = base64.b64encode(message_bytes)
    id: str = base64_bytes.decode('utf-8')
    return id


def decode(id: str) -> str:
    base64_bytes: bytes = id.encode('utf-8')
    message_bytes: bytes = base64.b64decode(base64_bytes)
    name: str = message_bytes.decode('utf-8')
    return name



# Copyright (C) 2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from typing import Protocol


class SqlExecutor(Protocol):
    def __call__(
            self,
            sql: str,
            data: tuple,
            do_commit: bool) -> int:
        ...


class SqlSelector(Protocol):
    def __call__(
            self,
            sql: str,
            parameters: tuple) -> list[any]:
        ...


def create_simple_insert_sql(
        table_name: str,
        column_list: list[str]) -> str:
    return f"""
        INSERT INTO
            {table_name}
            ({', '.join(column_list)})
        VALUES
            ({', '.join('?' * len(column_list))})
    """


def create_simple_delete_sql(
        table_name: str,
        where_column_list: list[str]) -> str:
    return f"""
        DELETE FROM
            {table_name}
        WHERE
            {" AND ".join(list(map(_append_question_mark, where_column_list)))}
    """


def create_simple_update_sql(
        table_name: str,
        set_column_list: list[str],
        where_column_list: list[str]) -> str:
    return f"""
        UPDATE
            {table_name}
        SET
            {', '.join(list(map(_append_question_mark, set_column_list)))}
        WHERE
            {' AND '.join(list(map(_append_question_mark, where_column_list)))}
    """


def create_simple_select_sql(
        table_name: str,
        select_column_list: list[str],
        where_column_list: list[str] = None) -> str:
    sql: str = f"""
        SELECT
            {', '.join(select_column_list)}
        FROM {table_name}
    """
    if where_column_list and len(where_column_list) > 0:
        sql = f"""
            {sql}
            WHERE
                {" AND ".join(list(map(_append_question_mark, where_column_list)))}
            """
    return sql


def create_simple_count_sql(table_name: str) -> str:
    return f"""
        SELECT COUNT(*) FROM
            {table_name}
    """


def _append_question_mark(column_name: str) -> str:
    return f"{column_name} = ?"


def neutral_execute_update(
        sql_executor: SqlExecutor,
        sql: str,
        data: tuple,
        do_commit: bool = True) -> int:
    return sql_executor(
        sql=sql,
        data=data,
        do_commit=do_commit)


def neutral_select(
        sql_selector: SqlSelector,
        sql: str,
        parameters: tuple) -> list[any]:
    sql_selector(
        sql=sql,
        parameters=parameters)



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class ElementType(Enum):
    TAG = 0, "tag"
    CATEGORY = 1, "cat",
    ALBUM = 2, "lbm",
    PLAYLIST = 3, "pl"
    MIX = 4, "mix"
    PAGELINK = 5, "pglnk"
    ARTIST = 6, "rtst"
    ARTIST_ALBUM_ALBUMS = 7, "rtstlbms"
    ARTIST_ALBUM_EP_SINGLES = 8, "rtstepsngls"
    ARTIST_ALBUM_OTHERS = 9, "rtstthrs"
    SIMILAR_ARTISTS = 10, "smlrrtsts"
    PLAYLIST_CONTAINER = 11, "plc"
    PLAYLIST_NAVIGABLE = 12, "pln"
    PLAYLIST_NAVIGABLE_ITEM = 13, "plni"
    MIX_CONTAINER = 14, "mixc"
    MIX_NAVIGABLE = 15, "mixn",
    MIX_NAVIGABLE_ITEM = 16, "mixi"
    ALBUM_CONTAINER = 17, "lbmc"
    ARTIST_TOP_TRACKS_NAVIGABLE = 18, "rtstttrks"
    NAVIGABLE_TRACK = 19, "trkn"
    ARTIST_RADIO_NAVIGABLE = 20, "rtstrd"
    ARTIST_TOP_TRACKS_LIST = 21, "rtstttrkslst"
    ARTIST_RADIO_LIST = 22, "rtstrdlst",
    RECENTLY_PLAYED_TRACKS_NAVIGABLE = 23, "rcntlptrks",
    RECENTLY_PLAYED_TRACKS_LIST = 24, "rcntlptrkslst",
    MOST_PLAYED_TRACKS_NAVIGABLE = 25, "mptrks",
    MOST_PLAYED_TRACKS_LIST = 26, "mptrkslst",
    FAVORITE_TRACKS_NAVIGABLE = 27, "fvrttrksn",
    FAVORITE_TRACKS_LIST = 28, "fvrttrksl",
    PAGE = 29, "pg",
    RECENTLY_PLAYED_ALBUMS = 30, "rcntpllbms",
    MOST_PLAYED_ALBUMS = 31, "mplbms",
    FAV_ARTIST_ADD = 32, "fvrtstadd",
    FAV_ARTIST_DEL = 33, "fvrtstdel",
    FAV_ALBUM_ADD = 34, "fvrtlbmadd",
    FAV_ALBUM_DEL = 35, "fvrtlbmdel",
    REMOVE_ALBUM_FROM_STATS = 36, "rmvlbmstts",
    REMOVE_TRACK_FROM_STATS = 37, "rmvtrkstts",
    FAVORITE_ALBUMS_BY_ARTIST_ASC = 39, "fvlbmbrtsta",
    FAVORITE_ALBUMS_BY_ARTIST_DESC = 40, "fvlbmbrtstd",
    FAVORITE_ALBUMS_BY_TITLE_ASC = 41, "fvlbmttla",
    FAVORITE_ALBUMS_BY_TITLE_DESC = 42, "fvlbmttld",
    FAVORITE_ALBUMS_BY_RELEASE_DATE_ASC = 43, "fvlbmrdta",
    FAVORITE_ALBUMS_BY_RELEASE_DATE_DESC = 44, "fvlbmrdtd",
    FAVORITE_ALBUMS_BY_USER_DATE_ADDED_ASC = 45, "fvlbmudaa",
    FAVORITE_ALBUMS_BY_USER_DATE_ADDED_DESC = 46, "fvlbmudad",
    FAVORITE_ARTISTS_BY_NAME_ASC = 47, "fvrtstna",
    FAVORITE_ARTISTS_BY_NAME_DESC = 48, "fvrtstnd",
    FAVORITE_ARTISTS_BY_USER_DATE_ADDED_ASC = 49, "fvrtstudaa",
    FAVORITE_ARTISTS_BY_USER_DATE_ADDED_DESC = 50, "fvrtstudad",
    ARTIST_FOCUS = 51, "rtstrl"
    ALBUMS_IN_MIX_OR_PLAYLIST = 52, "lbmnmxplst"
    ARTISTS_IN_MIX_OR_PLAYLIST = 53, "rtstnmxplst"
    BOOKMARK_ALBUM_ACTION = 54, "lbmlqctn"
    BOOKMARK_ALBUMS = 55, "lbllstnq"
    ALBUM_TRACKS = 56, "lbmtrks"
    TRACK_FAVORITE_ACTION = 57, "trkfvtct"
    BOOKMARK_ARTISTS = 58, "rtstlstnq"
    BOOKMARK_ARTIST_ACTION = 59, "rtstlqctn"
    BOOKMARK_TRACKS = 60, "snglstnq"
    BOOKMARK_TRACK_ACTION = 61, "snglstnqctn"
    MISSING_ALBUM = 62, "msnglbm"
    MISSING_TRACK = 63, "msngtrk"
    MISSING_ARTIST = 65, "msngrtst"
    ALL_TRACKS_IN_PLAYLIST_OR_MIX = 64, "lltrksmxplst"
    TRACK = 101, "trk"
    TRACK_CONTAINER = 102, "trkc"
    SEARCH = 103, "srch"

    def __init__(
            self,
            num: int,
            element_name: str):
        self.num: int = num
        self.element_name: str = element_name

    def getName(self):
        return self.element_name


def get_element_type_by_name(element_name: str) -> ElementType:
    # msgproc.log(f"get_tidal_tag_type_by_name with {tag_name}")
    for _, member in ElementType.__members__.items():
        if element_name == member.getName():
            return member
    raise Exception(f"get_element_type_by_name with {element_name} NOT found")


# duplicate check
name_checker_set: set[str] = set()
id_checker_set: set[int] = set()
for v in ElementType:
    if v.getName() in name_checker_set:
        raise Exception(f"Duplicated name [{v.getName()}]")
    if v.value[0] in id_checker_set:
        raise Exception(f"Duplicated id [{v.value[0]}]")
    name_checker_set.add(v.getName())
    id_checker_set.add(v.value[0])



#
# Copyright (C) 2017-2026 J.F.Dockes
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# [folders] section of the tree. This is the first part constructed and it is in charge of the
# initial fetching of data from recoll.
#
# Object Id prefix: 0$uprcl$folders
#
# Data structure:
#
# The _rcldocs list has one entry for each document in the index (mime:* search), it is the result
# of a 'mime:*' recoll query. Each entry is, or is similar to, a recoll.Doc(), a kind of dict.
#
# The _dirvec list has one entry for each directory. Each entry is a Python dict, mapping the
# directory entries' names to a pair (diridx,docidx), where:
#
#  - diridx is an index into _dirvec if the name is a directory, else -1
#  - docidx is an index into _rcldocs, or -1 if:
#     - There is no _rcldocs entry, which happens if there is no corresponding recoll doc for an
#       intermediary element in a path, or because this is a synthetic 'Group' entry.
#     - Or, while we build the structure, temporarily, if the doc was not yet seen. The value will
#       then be updated when we see it.
#
# Directories are created as needed by splitting the paths/urls from _rcldocs (and possibly adding
# some for groupings defined by the Group tag). Directories have no direct relation to their
# possible recoll index objects (if any), they are identified by their _dirvec index
#
# Note: docidx is usually set in the pair for a directory, if there is a Doc entry, but I don't
# think that it is ever used. The Recoll Doc for a directory has nothing very interesting in it.
#
# Note: We could probably use a single value, with a convention saying, e.g., that > 0 is for docs
# and < -1 for folders. Check if this saves a significant amount of memory.
#
# Each directory has a special ".." entry with a diridx pointing to the parent directory. This
# allows building a path from a container id (aka pwd).
#
# Only playlists have a "." entry (needed during init)
#
# Entry 0 in _dirvec is special: it holds the 'topdirs' from the recoll configuration. The entries
# are paths instead of simple names, and their docidx is 0. The diridx points to a regular dirvec
# entry.
#
# Object ids inside the section:
#    Container: $d<diridx> where <diridx> indexes into _dirvec
#    Item: $i<docidx> where <docidx> indexes into _rcldocs
# Note: this is very different from what Minim does. Minim uses actual objid paths as objids.
# E.g. 0$folders$f1589$f1593$f1604$*i11609. Must make pwd and any walk up the tree much easier.
# 
# We used to build an _xid2idx xdocid->objidx map to allow a Recoll item search result to be
# connected back to the folders tree, but this was not actually useful (bogus objids for items in
# search results are quite probably ok). Also quite probably, this could also be done using the URL,
# as it is what we use to build the folders tree in the first place. _xid2idx is currently
# desactivated (see comment). See objidfordoc() in this file for how we compute objids for
# search results which are actual directories (which need to be browsable).
#


import os
import sys
import time

from upmplgutils import uplog, direntry, getOptionValue
from uprclutils import audiomtypes, rcldoctoentry, cmpentries
import uprclutils
import uprclinit
import uprclfolderscreate


# All standard cover art file names:
_artexts = (".jpg", ".png")
def _artnamegen(base):
    for ext in _artexts:
        yield base + ext
_folderartbases = ("cover", "folder")
_folderartnames = []
for base in _folderartbases:
    for path in _artnamegen(base):
        _folderartnames.append(path)


# Create bogus "doc" for a path
def _docforpath(path, isdir=False):
    doc = {"url" : "file://" + path, "group": None, "embdimg" : None}
    if isdir:
        doc["mtype"] = "inode/directory"
    return doc


class Folders(object):

    # Initialize (read recoll data and build tree).
    def __init__(self, confdir, httphp, pathprefix):
        self._idprefix = "0$uprcl$folders"
        self._httphp = httphp
        self._pprefix = pathprefix
        self._rcldocs = uprclfolderscreate._fetchalldocs(confdir)
        # _playlists is used to store the diridx for the playlists during the initial walk, for
        # initialization when the tree is complete.
        self._dirvec, self._playlists = uprclfolderscreate._rcl2folders(confdir, self._rcldocs)
        # _moredocs is overflow storage for synthetic records created for playlists url
        # entries. Uses docidx values starting at len(_rcldocs), with index into moredocs
        # (value - len(_rcldocs))
        self._moredocs = uprclfolderscreate._initplaylists(self, 
            confdir, self._rcldocs, self._dirvec, self._playlists)
        self._enabletags = uprclinit.g_minimconfig.getboolvalue("showExtras", True)
        self._notagview = getOptionValue("uprclnotagview", False)


    def rcldocs(self):
        return self._rcldocs


    # Tell the top module what entries we define in the root
    def rootentries(self, pid):
        return [
            direntry(pid + "folders", pid, "[folders]"),
        ]


    # Given a doc, walk its url down from the part in root to find its directory entry, and
    # return the entry's_dirvec and _rcldocs indices, either of which can be -1
    def _stat(self, doc, verbose=False):
        # _splitpath returns the _dirvec index of the topdirs entry in root that we start from and
        # the split rest of path.  That is if the doc url has /av/mp3/classique/bach/ and the root
        # entry is /av/mp3, we get the _dirvec entry index for /av/mp3 and [classique, bach]
        # _splitpath is in the creation module because it is used before our _dirvec variable is
        # set.
        fathidx, pathl = uprclfolderscreate._splitpath(self._dirvec, doc)
        if verbose:
            uplog(f"_stat: pbtd returns fathidx {fathidx} pathl {pathl}")
        if not fathidx:
            return -1, -1
        docidx = -1
        for elt in pathl:
            if not elt in self._dirvec[fathidx]:
                if verbose:
                    uplog(f"_stat: element [{elt}] has no entry in {fathidx} "
                          "[{self._dirvec[fathidx]}.keys()]")
                return -1, -1
            if verbose:
                uplog(f"_stat: element [{elt}] entry in {fathidx} [{self._dirvec[fathidx][elt]}]")
            fathidx, docidx = self._dirvec[fathidx][elt]

        return fathidx, docidx


    # Find the doc index for a filesystem path.
    # We use a temporary doc to call _stat()
    def statpath(self, path, verbose=False):
        doc = _docforpath(path)
        fathidx, docidx = self._stat(doc, verbose)
        if docidx >= 0 and docidx < len(self._rcldocs):
            return docidx
        return -1


    # Extract diridx and further path (leading to tags) from objid.
    def _objidtoidx(self, pid):
        if not pid.startswith(self._idprefix):
            raise Exception(f"folders.browse: bad pid [{pid}]")

        if len(self._rcldocs) == 0:
            raise Exception("folders:browse: no docs")

        isitem = False
        dirpth = pid[len(self._idprefix) :]
        if not dirpth:
            idx = 0
            pathremain = ""
        else:
            if dirpth[0:2] == "$d":
                isitem = False
            elif dirpth[0:2] == "$i":
                isitem = True
            else:
                raise Exception("folders:browse: called on non dir objid %s" % pid)
            # Other $sign?
            nextdol = dirpth.find("$", 1)
            if nextdol > 0:
                idx = int(dirpth[2:nextdol])
            else:
                idx = int(dirpth[2:])
            pathremain = None
            if nextdol > 0:
                pathremain = dirpth[nextdol:]

        if isitem:
            if idx >= len(self._rcldocs):
                raise Exception(f"folders:browse: bad pid exceeds rcldocs size [{pid}]")
        else:
            if idx >= len(self._dirvec):
                raise Exception(f"folders:browse: bad pid exceeds dirvec size [{pid}]")

        return (isitem, idx, pathremain)


    def _docforidx(self, docidx):
        if docidx < len(self._rcldocs):
            return self._rcldocs[docidx]
        else:
            idx = docidx - len(self._rcldocs)
            if idx < len(self._moredocs):
                return self._moredocs[idx]
        return None


    # Look all non-directory docs inside directory, and return the cover art we find.
    # 
    # TBD:
    # - In the case where this is a Group directory, we'd need to go look into the file system for a
    #   group.xxx image.  As it is, things work if the tracks rely on the group pic (instead of
    #   having an embedded pic or track pic)
    # - playlists: need to look at the physical dir for a e.g. playlistname.jpg.
    #
    # We used to only look for art for direct children tracks, but we now also look at
    # subdirs. This will yield an image from the first subdir which has an image file in
    # it, so somewhat random, but nice anyway.
    def _arturifordironedoc(self, diridx, docidx):
        # Look for art for one object, track or directory.
        # Directories only have doc entries if they are also albums. Else we fake a doc.
        if docidx >= 0 and docidx < len(self._rcldocs):
            doc = self._rcldocs[docidx]
        else:
            doc = _docforpath(self.dirpath("", diridx), True)
        return self.docarturi(doc, preferfolder=True)

    # Look for art for the directory itself, then its children.
    def _arturifordir(self, thisdiridx, thisdocidx=-1):
        # First look at the directory itself.
        arturi = self._arturifordironedoc(thisdiridx, thisdocidx)
        if arturi:
            return arturi
        # Then look at children.
        for nm, ids in self._dirvec[thisdiridx].items():
            diridx = ids[0]
            docidx = ids[1]
            arturi = self._arturifordironedoc(diridx, docidx)
            if arturi:
                return arturi


    def _browsemeta(self, pid, isitem, idx):
        docidx = -1
        if isitem:
            docidx = idx
        else:
            try:
                ids = self._dirvec[idx]
                docidx = ids[1]
            except:
                pass
        if docidx != -1:
            doc = self._docforidx(docidx)
            id = self._idprefix + "$i" + str(docidx)
            e = rcldoctoentry(id, pid, self._httphp, self._pprefix, doc)
            return [e,]


    ##############
    # Browsing the initialized [folders] hierarchy
    # @param pid objid is like folders$index
    # @param flag: "meta" or "children".
    def browse(self, pid, flag, offset, count):

        isitem, idx, pthremain = self._objidtoidx(pid)

        # If pthremain is set, this is pointing to 'Tag View'. Pass
        # the request along to the tags browser.
        if pthremain:
            if not pthremain.find("$tagview.0") == 0:
                raise Exception(f"uprclfolders:browse: pid [{pid}]. bad pthremain")
            return uprclinit.getTree("tags").browseFolder(pid, flag, pthremain, self.dirpath(pid))

        # If there is only one entry in root, skip it. This means that 0 and 1 point to the same
        # dir, but this does not seem to be an issue
        if not isitem and idx == 0 and len(self._dirvec[0]) == 2:
            idx = 1

        if flag == "meta":
            if not isitem:
                raise Exception(f"uprclfolders:browse: browsemeta on non-item pid [{pid}]")
            return self._browsemeta(pid, isitem, idx)

        # uplog(f"Folders browse: idx [{idx}] content: [{self._dirvec[idx]}]")
        entries = []
        showtopart = True
        # The basename call is just for diridx==0 (topdirs). Remove it if
        # this proves a performance issue
        
        for nm, ids in self._dirvec[idx].items():
            if nm == ".." or nm == ".":
                continue
            thisdiridx = ids[0]
            thisdocidx = ids[1]
            if thisdiridx >= 0:
                # Skip empty directories
                if len(self._dirvec[thisdiridx]) == 1:
                    continue
                # If there are directories, don't show art for the Tags top entries, this would
                # show one of the subdir's art and looks weird
                showtopart = False
                id = self._idprefix + "$d" + str(thisdiridx)
                arturi = self._arturifordir(thisdiridx, thisdocidx)
                entries.append(direntry(id, pid, os.path.basename(nm), arturi=arturi))
            else:
                # Not a directory. docidx had better been set
                if thisdocidx == -1:
                    uplog("folders:docidx -1 for non-dir entry %s" % nm)
                    continue
                doc = self._docforidx(thisdocidx)

                id = self._idprefix + "$i" + str(thisdocidx)
                e = rcldoctoentry(id, pid, self._httphp, self._pprefix, doc)
                if e:
                    entries.append(e)

        if idx not in self._playlists:
            entries.sort(key=cmpentries)

        # Add "Browse subtree by tags" entry
        if (not self._notagview) and pid != self._idprefix and self._enabletags:
            arturi = None
            if showtopart:
                arturi = self._arturifordir(idx)
            id = pid + "$tagview.0"
            entries.insert(0, direntry(id, pid, ">> Tag View", arturi=arturi))

        return entries


    # Return path for objid, which has to be a container.This is good old
    # pwd... It is called from the search module for generating a 'dir:'
    # recoll filtering directive.
    def dirpath(self, objid, diridx=None):
        # We may get called from search, on the top dir (above [folders]). Return empty in this case
        if not diridx:
            try:
                isitem, diridx, pthremain = self._objidtoidx(objid)
            except:
                return ""
            if isitem:
                raise Exception(f"uprclfolders:dirpath: called on item pid [{pid}]")

        if diridx == 0:
            return "/"

        lpath = []
        while True:
            fathidx = self._dirvec[diridx][".."][0]
            found = False
            for nm, ids in self._dirvec[fathidx].items():
                if ids[0] == diridx:
                    lpath.append(nm)
                    found = True
                    break
            # End for
            if not found:
                uplog(f"uprclfolders: pwd failed for {objid} (father not found), returning /")
                return "/"
            if len(lpath) > 200:
                uplog(f"uprclfolders: pwd failed for {objid} (looping), returning /")
                return "/"

            diridx = fathidx
            if diridx == 0:
                break

        if not lpath:
            path = "/"
        else:
            path = ""
        for elt in reversed(lpath):
            path += elt + "/"

        return path


    # Compute an UPnP object ID for a document produced by a recoll search
    def objidfordoc(self, doc):
        id = None
        if doc["mtype"] == "inode/directory":
            fathidx, docidx = self._stat(doc)
            return self._idprefix + "$d" + str(fathidx)
        else:
            # Note: we thought we should have something like objidforxdocid for using consistent
            # objids, but it does not seem to be actually necessary. Use a unique but different id
            # instead.
            id = self._idprefix + "$xdocid" + doc.xdocid
        # uplog(f"objidfordoc: returning {id}")
        return id


    # Track-specific cover art. Based on either the file name or embedded data
    def _trackarturi(self, doc, objpath):
        # Check for an image specific to the track file
        base, ext = os.path.splitext(objpath)
        for artpath in _artnamegen(base):
            #uplog(f"_trackarturi: checking existence:[{artpath}]")
            artdoc = _docforpath(artpath)
            fathidx, docidx = self._stat(artdoc)
            if docidx >= 0:
                return uprclutils.httpurl(self._httphp, os.path.join(self._pprefix, artpath))

        # Else try to use an embedded img
        if doc["embdimg"]:
            arturi = uprclutils.embdimgurl(doc, self._httphp, self._pprefix)
            if arturi:
                # uplog("docarturi: embedded: %s"%printable(arturi))
                return arturi
        return None


    # Return folder-level art uri (e.g. /path/to/folder.jpg) if it exists
    def _folderart(self, doc, albtitle=None):
        # If doc is a directory, this returns it own path, else the father path
        folderpath = uprclutils.docfolder(doc)
        dirdoc = _docforpath(folderpath, True)
        folderidx, _ = self._stat(dirdoc)
        if folderidx < 0:
            uplog(f"_folderart: folder not found: {folderpath}")
            return None

        foldercontents = self._dirvec[folderidx].keys()
        #uplog(f"_folderart: path [{folderpath}] idx {folderidx} contents [{foldercontents}]")

        # If albtitle is set check for an image of the same name
        if albtitle:
            for fsimple in _artnamegen(albtitle):
                if fsimple in foldercontents:
                    return uprclutils.httpurl(self._httphp, os.path.join(folderpath, fsimple))

        # Look for an appropriate image in the file folder. We list the folder and look for a
        # case-insensitive match for all the possible cover art conventional names
        arturi = None
        for f in foldercontents:
            flowersimple = f.lower()
            if flowersimple in _folderartnames:
                path = os.path.join(self._pprefix, folderpath, f)
                arturi = uprclutils.httpurl(self._httphp, path)
                break

        #uplog(f"folderart: returning {arturi}")
        return arturi


    # Find cover art for doc which may be a folder or track. We can look for both a
    # track-specific image or a folder-level one
    #
    # We return a special uri if the file has embedded image data
    def docarturi(self, doc, preferfolder=False, albtitle=None):
        objpath = doc["url"][7:]
        #uplog(f"docarturi: preferfolder {preferfolder} docpath {objpath}")
        
        if not preferfolder:
            arturi = self._trackarturi(doc, objpath)
            if arturi:
                return arturi
        
        # won't work for the virtual group directory itself: it has no doc
        if doc["group"]:
            base = os.path.join(os.path.dirname(objpath), uprclutils.tag2fn(doc["group"]))
            for artpath in _artnamegen(base):
                #uplog(f"docarturi:calling os.path.exist({artpath})")
                if os.path.exists(artpath):
                    return uprclutils.httpurl(self._httphp, os.path.join(self._pprefix, artpath))
        
        # TBD Here minimserver would look for album disc before album art (which is taken care of by
        # _folderart() with albtitle set)
        # Look for folder level image file (e.g. cover.jpg)
        arturi = self._folderart(doc, albtitle)
        if arturi:
            return arturi
        
        # If preferfolder is set, we did not look at the track-specific art, do it last.
        if preferfolder:
            arturi = self._trackarturi(doc, objpath)
        
        return arturi



# Copyright (C) 2023 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

class RadioStationEntry:

    def __init__(
            self,
            id : int,
            codec : str,
            url : str,
            title : str,
            mimetype : str,
            bit_depth : int,
            sampling_rate : int,
            channel_count : int):
        self._id : int = id
        self._title : str = title
        self._url : str = url
        self._codec : str = codec
        self._mimetype : str = mimetype
        self._bit_depth : int = bit_depth
        self._sampling_rate : int = sampling_rate
        self._channel_count : int = channel_count

    @property
    def id(self) -> int:
        return self._id

    @property
    def title(self) -> str:
        return self._title

    @property
    def url(self) -> str:
        return self._url
    
    @property
    def codec(self) -> str:
        return self._codec

    @property
    def mimetype(self) -> str:
        return self._mimetype

    @property
    def bit_depth(self) -> int:
        return self._bit_depth

    @property
    def sampling_rate(self) -> int:
        return self._sampling_rate

    @property
    def channel_count(self) -> int:
        return self._channel_count



# Copyright (C) 2017-2022 J.F.Dockes
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as published by
#   the Free Software Foundation; either version 2.1 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with this program; if not, write to the
#   Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import os
import time
import re
from recoll import recoll

from uprclutils import audiomtypes, docfolder, uplog
import uprclutils
import uprclinit

# Tags for which we may create auxiliary tag tables for facet descent.
#
# The dict key is the tag name, used as container title in the tags tree,
# except if an a different display value was set by the user
# (_tagdisplaytag). The value is the auxiliary table name, used also
# as base for unique id and join columns (with _id) appended, and is
# also currently the recoll field name (with a provision to differ if
# needed, thanks to the currently empty _coltorclfield dict).
#
# The keys must stay same as what Minim uses as we filter them with indexTags from its
# configuration.
#
# Note: artist is actually doc["albumartist"] if set, else doc["artist"].
#
# This is the base set, with some custom translations (e.g. Group->cgroup). If a minimserver
# configuration is used, Some entries may be filtered out and some may be added.
_alltagtotable = {
    "AlbumArtist": "albumartist",
    "All Artists": "allartists",
    "Artist": "artist",  # Special processing in recolltosql: "albumartist" if set else "artist"
    "Comment": "comment",
    "Composer": "composer",
    "Conductor": "conductor",
    "Date": "date",
    "Genre": "genre",
    "Group": "cgroup",  # can't have a table named group
    "Label": "label",
    "Lyricist": "lyricist",
    "Orchestra": "orchestra",
    "Performer": "performer",
}

# Translations used when fetching fields from the recoll record. Most have the same name as the
# column.
_coltorclfield = {"allartists": "artist"}

# The actual list of tags (e.g. "Genre", "Composer") which will be shown in the tree, after reading
# the local configuration.
g_indextags = []
# The display string for the above tags. Non-identity if there are entries like Composer:Compositeur
# in the indexTags minim config
g_tagdisplaytag = {}
# Same as _alltagtotable, filtered by g_indextags
g_tagtotable = {}


def getIndexTags():
    return g_indextags


def getTagDisplayTag():
    return g_tagdisplaytag


def getTagToTable():
    return g_tagtotable


# Name of the join column for one of the tag tables. Actually, now that we use separate junction
# tables instead of columns inside the tracks table, the column name could be constant. Kept it.
def _clid(table):
    return table + "_id"


# Name of the junction table for a given tag table. Ex for genres.
# rcldocs is not an sqlite table, it's the recoll document list.
#         docidx        docidx                genre_id
# rcldocs<------>tracks<------>tracks_genres<---------->genre
def _junctb(table):
    return "tracks_" + table + "s"


# Create an empty db.
#
# There is one table for each tag (Artist, Genre, Date, etc.), listing all possible values for the
# particular tag. The table name is the value from the _alltagtotable array, <tagname> and it has 2
# columns: <tagname>_id and value.
#
# The tracks table is the "main" table, and has a record for each track, with a title column, and an
# album join column (album_id) (because a track can only belong to one album). The unique index into
# this table is the document index in the recoll document list (docidx).
#
# Because a track can have multiple values for a given tag, we use junction tables. These are named
# tracks_<tagname>s (e.g. tracks_genres). The records in these have two columns: docidx (track id)
# and and <tagname>_id (unique id from the <tagname> table)
#
# The Albums table is special because it is built according to, and stores, the file system location
# (the album title is not enough to group tracks, there could be many albums with the same
# title). Also we do tricks with grouping discs into albums etc.
#
# Note: we create all tables even if not all tags are actually used.
def _createsqdb(conn):
    c = conn.cursor()

    # Create the albums table
    try:
        c.execute("""DROP TABLE albums""")
    except:
        pass
    c.execute(
        "CREATE TABLE albums (album_id INTEGER PRIMARY KEY,"
        # artist_id is for an album_artist, either explicit from an albumartist attribute or
        # implicit if all tracks have the same artist attribute
        "artist_id INT,"
        "albtitle TEXT, albfolder TEXT, albdate TEXT, albarturi TEXT,"
        # During creation, albalb is non null only the elements of a merged album, and is the
        # album_id id for the parent. It is set by the album merging pass. When the album merge is
        # done, albalb is set to album_id for all remaining (single-disc, non merged) albums.
        "albalb INT,"
        # albtdisc is set from the tracks DISCNUMBER tag or equivalent, while initially walking the
        # docs. It signals candidates for merging. album/discs with albtdisc set are not selected
        # for display in album lists, as they are supposedly represented by their parent album. The
        # field is reset to null for albums with a discnumber which do not end up being merged.
        "albtdisc INT,"
        # Temp column while creating the table for deciding if we have a uniform artist or not. This
        # is a text version of all the artistid sets for each album track. If we were really
        # concerned about space usage, we could use a parallel temp table and drop it when done...
        "artists TEXT"
        ")"
    )

    # Create the main tracks table, which has a one-to-one relationship with the recoll document
    # array
    try:
        c.execute("""DROP TABLE tracks""")
    except:
        pass
    tracksstmt = """CREATE TABLE tracks 
                     (docidx INT, album_id INT, trackno INT, title TEXT, path TEXT)"""
    c.execute(tracksstmt)

    # Create tables for tag values (e.g. all genre values, all composer values, etc.)
    for tb in _alltagtotable.values():
        _createTagTables(c, tb)


# Create the value and junction table for a given tag
def _createTagTables(cursor, tgnm):
    # Create a table for this tag values and value ids
    try:
        cursor.execute("DROP TABLE " + tgnm)
    except:
        pass
    stmt = "CREATE TABLE " + tgnm + " (" + _clid(tgnm) + " INTEGER PRIMARY KEY, value TEXT)"
    cursor.execute(stmt)
    # Create a junction table between the tracks table and the tag values table, with docidx and
    # valueid columns, allowing multiple tag values for a given track (e.g. several
    # genres). Using a table-specific name for the id column would not be necessary, it's a
    # remnant of the time where these columns were integral to the tracks table. Still, does not
    # hurt.
    try:
        cursor.execute("DROP TABLE " + _junctb(tgnm))
    except:
        pass
    stmt = f"CREATE TABLE {_junctb(tgnm)} (docidx INT, {_clid(tgnm)} INT, " \
        f"UNIQUE(docidx, {_clid(tgnm)}))"
    cursor.execute(stmt)


# Augment indextag dict and create tables for a custom field, not part of our predefined set
def _addCustomTable(conn, indextag):
    c = conn.cursor()
    tb = indextag.lower()
    _alltagtotable[indextag] = tb
    _createTagTables(c, tb)


# Peruse the configuration to decide what tags will actually show up
# in the tree and how they will be displayed.
def _prepareTags(conn):
    global g_tagdisplaytag
    global g_tagtotable
    global g_indextags

    g_indextags = []
    g_tagdisplaytag = {}
    g_tagtotable = {}

    indextagsp = uprclinit.g_minimconfig.getindextags()
    itemtags = uprclinit.g_minimconfig.getitemtags()
    if not indextagsp:
        # Our default list of index tags (showing up as containers in the tree). The second field in
        # each pair, if set, is the customized display (e.g. ("Composer", "Compositeur"))
        indextagsp = [("Artist", ""), ("Date", ""), ("Genre", ""), ("Composer", "")]

    # Compute the actual list of index tags:
    for v, d in indextagsp:
        if v.lower() == "none":
            g_indextags = []
            g_tagdisplaytag = {}
            break
        g_indextags.append(v)
        g_tagdisplaytag[v] = d if d else v
    uplog("prepareTags: g_indextags: %s g_tagdisplaytag %s" % (g_indextags, g_tagdisplaytag))

    # Compute an array of (table name, recoll field) translations for the tags we need to process,
    # as determined by the indexTags property. Most often they are identical. This also determines
    # what fields we create tables for.
    tabtorclfield = []
    for nm in g_indextags:
        if nm not in _alltagtotable:
            _addCustomTable(conn, nm)
        tb = _alltagtotable[nm]
        g_tagtotable[nm] = tb
        rclfld = _coltorclfield[tb] if tb in _coltorclfield else tb
        uplog(f"prepareTags: using rclfield [{rclfld}] for sql [{tb}]")
        tabtorclfield.append((tb, rclfld))
        # Special case "artist" because we also want "composer" and "conductor" in there.  This is
        # so that recoll doc records will be created by _artiststorecoll(), and search for artist by
        # title will find them. This is a bit dubious because, if there are Composer or Conductor
        # indexes, we are duplicating data, and the user probably does not want to find composers in
        # the artist list. The other possible approach would (tbchecked?) be to use itemtags tables
        # but this would need additional code in uprcltags.py to duplicate the direntryforartid()
        # method and the code in uprclsearch which calls it + maybe some of the code which links
        # artists to albums and tracks probably? may worth a try anyway, it would be nicer if not
        # too complicated.
        if tb == "artist":
            tabtorclfield.append((tb, "composer"))
            tabtorclfield.append((tb, "conductor"))

    for nm in itemtags:
        if nm not in _alltagtotable:
            _addCustomTable(conn, nm)
        tb = _alltagtotable[nm]
        rclfld = _coltorclfield[tb] if tb in _coltorclfield else tb
        uplog(f"prepareTags: using rclfield [{rclfld}] for sql [{tb}]")
        tabtorclfield.append((tb, rclfld))

    uplog(f"prepareTags: tabtorclfield: {tabtorclfield}")
    return tabtorclfield


# Insert new value if not existing, return rowid of new or existing row
def _auxtableinsert(conn, tb, value):
    # uplog(f"_auxtableinsert {tb} -> {value}")
    c = conn.cursor()
    stmt = f"SELECT {_clid(tb)} FROM  {tb} WHERE value = ?"
    c.execute(stmt, (value,))
    r = c.fetchone()
    if r:
        rowid = r[0]
    else:
        stmt = f"INSERT INTO {tb}(value) VALUES(?)"
        c.execute(stmt, (value,))
        rowid = c.lastrowid
    return rowid


# tracknos like n/max are now supposedly processed by rclaudio and
# should not arrive here, but let's play it safe.
def _tracknofordoc(doc):
    try:
        return int(doc["tracknumber"].split("/")[0])
    except:
        # uplog("_tracknofordoc: doc["tracknumber"] %s title %s url %s" %
        #      (doc["tracknumber"], doc["title"], doc["url"]))
        return 1


# Detect '[disc N]' or '(disc N)' or ' disc N' at the end of an album title
# Groups         1   2       3                 4                      5
_albtitdnumre = r"(.*)(\[disc ([0-9]+)\]|\(disc ([0-9]+)\)|[ ]+disc[ ]+([0-9]+))[ ]*$"
_albtitdnumexp = re.compile(_albtitdnumre, flags=re.IGNORECASE)
_folderdnumre = r"(cd|disc)[ ]*([0-9]+)(.*)"
_folderdnumexp = re.compile(_folderdnumre, flags=re.IGNORECASE)


# Create album record for track if needed (not already there).
# The albums table is special, can't use auxtableinsert()
def _maybecreatealbum(conn, doc):
    c = conn.cursor()
    folder = docfolder(doc)

    album = doc["album"]
    if not album:
        album = uprclutils.basename(folder)
        # uplog("Using %s for alb MIME %s title %s" % (album,doc["mtype"],doc["url"]))

    # uplog(f"_maybecreatealbum: album: [{album}] folder [{folder}]")

    if doc["albumartist"]:
        albartist_id = _auxtableinsert(conn, "artist", doc["albumartist"])
    else:
        albartist_id = None

    # See if there is a discnum, either explicit or from album
    # title
    discnum = None
    if doc["discnumber"]:
        try:
            discnum = int(doc["discnumber"])
            # uplog(f"discnumber {discnum} folder {folder}")
        except:
            pass

    # Look for a disc number at the end of the album title. If it's
    # there and it's the same as the discnumber attribute, we fix the
    # title, else we leave it alone (and the merge will later fail for
    # differing titles)
    m = _albtitdnumexp.search(album)
    if m:
        # uplog("Disc number found at end of album title")
        for i in (3, 4, 5):
            if m.group(i):
                adiscnum = int(m.group(i))
                # uplog(f"Match is in group {i} adiscnum {adiscnum} discnum {discnum}")
                if not discnum:
                    discnum = adiscnum
                if adiscnum == discnum:
                    album = m.group(1)
                    # uplog(f"album title changed to [{album}]")
                break

    if not discnum:
        m = _folderdnumexp.search(uprclutils.basename(folder))
        if m:
            discnum = int(m.group(2))

    # See if this albumdisc already exists (created for a previous track)
    stmt = "SELECT album_id, artist_id FROM albums WHERE albtitle = ? AND albfolder = ?"
    wcols = [album, folder]
    if discnum:
        stmt += " AND albtdisc = ?"
        wcols.append(discnum)
    # uplog(f"maybecreatealbum: {stmt} {wcols}")
    c.execute(stmt, wcols)
    r = c.fetchone()
    if r:
        # uplog("maybecreatealbum: album found")
        album_id = r[0]
    else:
        c.execute(
            """INSERT INTO 
        albums(albtitle, albfolder, artist_id, albdate, albtdisc, artists) 
        VALUES (?,?,?,?,?,?)""",
            (album, folder, albartist_id, doc["date"], discnum, ""),
        )
        album_id = c.lastrowid
        # uplog(f"New album {album_id} {album} disc {discnum} artist {albartist_id} folder {folder}")

    return album_id


# Add a track's artists set to the album auxiliary "artists" column, used in the end to determine an
# album artist if none was explicitely set.
# The contents of the column is a string made of string representations of Python sets of
# ints. E.g. "|{1234, 12345}|{99}|{23, 24}"
# At the end (all tracks seen), we reconstruct the sets and intersect them. If the intersection is
# not empty, we make a random element inside it the album artist. See _setalbumartists()
# It seems that we could accomplish the same thing by querying the db for each of the album tracks
# artists at the end ? I can't see the reason for this approach any more, but maybe there is one...
def _updatealbartistlist(conn, album_id, rowids):
    if not rowids:
        return
    c = conn.cursor()
    addstr = repr(rowids)
    stmt = f"UPDATE albums SET artists = artists || '|' || ? WHERE album_id = ?"
    values = (addstr, album_id)
    c.execute(stmt, values)


# Setting album covers needs to wait until we have scanned all tracks so that we can select
# a consistant embedded art (first track in path order), as recoll scanning is in unsorted
# directory order.
def _setalbumcovers(conn, folders):
    rcldocs = folders.rcldocs()
    c = conn.cursor()
    c.execute("""SELECT album_id,albtitle FROM albums""")
    for r in c:
        albid = r[0]
        albtitle = r[1]
        c1 = conn.cursor()
        stmt = """SELECT docidx FROM tracks WHERE album_id = ? ORDER BY path"""
        c1.execute(stmt, (albid,))
        for r1 in c1:
            docidx = r1[0]
            doc = rcldocs[docidx]
            arturi = folders.docarturi(
                doc,
                preferfolder=True,
                albtitle=albtitle,
            )
            if arturi:
                cupd = conn.cursor()
                # uplog(f"Setting albid {albid} albarturi to {arturi}")
                cupd.execute("UPDATE albums SET albarturi = ?  WHERE album_id = ?", (arturi, albid))
                break


# After the pass on tracks, look for all albums where the album artist is not set but all the tracks
# have the same artist, and set the album artist.
def _setalbumartists(conn):
    c = conn.cursor()
    # Get all albums without an albumartist where the same_track_artist flag is on
    c.execute("SELECT album_id, artists, albtitle FROM albums WHERE artist_id IS NULL")
    # For all albumartist-less albums
    for r in c:
        # Default to various artist
        albumartist = variousartistsid
        if r[1]:
            # Eval the string back to list of artist sets, and intersect to see if we have a common
            # artist
            s = r[1].lstrip("|")
            artsets = s.split("|")
            artsets = [eval(artset) for artset in artsets]
            if artsets:
                intersect = artsets[0]
                for artset in artsets[1:]:
                    intersect = intersect.intersection(artset)
                # uplog(f"_setalbumartists: INTERSECT for [{r[2]}]: {intersect}")
                if intersect:
                    # if multiple values, have to chose one...
                    albumartist = [v for v in intersect][0]
                    # uplog(f"Using albumartist {albumartist} for {r[2]}")

        if albumartist == variousartistsid:
            # uplog(f"Using albumartist Various Artists for {r[2]}")
            pass
        c1 = conn.cursor()
        c1.execute("UPDATE albums SET artist_id = ? WHERE album_id = ?", (albumartist, r[0]))


# Add albums to the recoll index so they can be searched for
def _albumstorecoll(conn, rcldb):
    c = conn.cursor()
    #                 0        1          2          3         4         5
    stmt = """SELECT album_id, albfolder, albtitle, albarturi, albdate, artist.value
      FROM albums LEFT JOIN artist ON artist.artist_id = albums.artist_id
      WHERE albtdisc is NULL"""
    c.execute(stmt, ())
    for r in c:
        udi = "albid" + str(r[0])
        doc = rcldb.doc()
        doc["title"] = r[2]
        doc["mtype"] = "inode/directory"
        if r[5]:
            doc["albumartist"] = r[5]
            doc["artist"] = r[5]
        doc["url"] = "file://" + r[1]
        # uplog(f"_albumstorecoll: indexing album {doc['title']}")
        rcldb.addOrUpdate(udi, doc)


# Add artists to the recoll index so they can be searched for
def _artiststorecoll(conn, rcldb):
    c = conn.cursor()
    #                 0         1
    stmt = "SELECT artist_id, value    FROM artist"
    c.execute(stmt, ())
    for r in c:
        udi = "artid" + str(r[0])
        doc = rcldb.doc()
        doc["title"] = r[1]
        doc["mtype"] = "inode/directory"
        doc["url"] = "file://artists/" + r[1]  # Not used ever
        # uplog(f"_artiststorecoll: indexing artist {doc['title']}")
        rcldb.addOrUpdate(udi, doc)


# Check that the numbers are sequential
def _checkseq(seq):
    num = seq[0]
    if not num:
        return False
    for e in seq[1:]:
        if e != num + 1:
            return False
        num = e
    return True


def _membertotopalbum(conn, memberalbid):
    c = conn.cursor()
    c.execute("""SELECT * FROM albums WHERE album_id = ?""", (memberalbid,))
    cols = [desc[0] for desc in c.description]
    # Get array of column values, and set primary key and albtdisc to
    # None before inserting the copy
    tdiscindex = cols.index("albtdisc")
    v = [e for e in c.fetchone()]
    v[0] = None
    v[tdiscindex] = None
    c.execute("""INSERT INTO albums VALUES (%s)""" % ",".join("?" * len(v)), v)
    return c.lastrowid


# Only keep albums in the same folder or in a sibling folder
def _mergealbumsfilterfolders(folder, rows, colidx):
    parent = uprclutils.dirname(folder)
    rows1 = []
    for row in rows:
        if row[colidx] == folder or uprclutils.dirname(row[colidx]) == parent:
            rows1.append(row)
    return rows1


## TBD: folder match filter
def _createmergedalbums(conn):
    c = conn.cursor()

    # Remember already merged
    merged = set()

    # All candidates for merging: albums with a disc number not yet
    # merged (albalb is null)
    c.execute(
        """SELECT album_id, albtitle, artist_id, albfolder FROM albums
      WHERE albalb IS NULL AND albtdisc IS NOT NULL"""
    )
    c1 = conn.cursor()
    for r in c:
        albid = r[0]
        if albid in merged:
            continue
        albtitle = r[1]
        artist = r[2]
        folder = r[3]

        # uplog(f"_createmergedalbums: albid {albid} artist_id {artist} albtitle {albtitle}")

        # Look for albums not already in a group, with the same title and artist
        if artist:
            c1.execute(
                """SELECT album_id, albtdisc, albfolder FROM albums
                WHERE albtitle = ? AND artist_id = ?
                AND albalb is NULL AND albtdisc IS NOT NULL""",
                (albtitle, artist),
            )
        else:
            c1.execute(
                """SELECT album_id, albtdisc, albfolder FROM albums
                WHERE albtitle = ? AND artist_id IS NULL
                AND albalb is NULL AND albtdisc IS NOT NULL""",
                (albtitle,),
            )

        rows = c1.fetchall()
        rows1 = _mergealbumsfilterfolders(folder, rows, 2)

        # uplog(f"_createmergedalbums: got {len(rows1)} possible(s) title {albtitle}")

        if len(rows1) > 1:
            albids = [row[0] for row in rows1]
            dnos = sorted([row[1] for row in rows1])
            if not _checkseq(dnos):
                uplog(f"mergealbums: not merging bad seq {dnos} for albtitle {albtitle}")
                c1.execute(
                    f"UPDATE albums SET albtdisc = NULL "
                    f"WHERE album_id in ({','.join('?'*len(albids))})",
                    albids,
                )
                continue

            # Create record for whole album by copying the first
            # record, setting its album_id and albtdisc to NULL
            topalbid = _membertotopalbum(conn, albids[0])

            # Update all album disc members with the top album id
            values = [
                topalbid,
            ] + albids
            c1.execute(
                f"UPDATE albums SET albalb = ? " f"WHERE album_id in ({','.join('?'*len(albids))})",
                values,
            )
            merged.update(albids)
            # uplog("_createmergedalbums: merged: %s" % albids)

        elif len(rows1) == 1:
            # Album with a single disc having a discnumber. Just unset
            # the discnumber, we won't use it and its presence would
            # prevent the album from showing up. Alternatively we
            # could set albalb = album_id?
            # uplog("Setting albtdisc to NULL albid %d" % albid)
            c1.execute("UPDATE albums SET albtdisc = NULL WHERE album_id= ?", (albid,))

    # finally, set albalb to albid for all single-disc albums
    c.execute("UPDATE albums SET albalb = album_id WHERE albtdisc IS NULL")


# Parse date: partly duplicated from rclaudio, see comments about date formats there.  The date
# field is a bit of a mess because up to some point recoll had an alias from date to dmtime, which
# resulted in no date field and a calendar date (instead of uxtime) in dmtime. So we take the date
# value from 'date' if set, else 'dmtime', and then process it here. All reasonable date formats
# begin with YYYY[-MM[-DD]] and this is the part we keep.
def parsedate(dt):
    if len(dt) > 10:
        dt = dt[0:10]
    l = dt.split("-")
    if len(l) > 3 or len(l) == 2 or len(l[0]) != 4 or l[0] == "0000":
        return None
    # uplog("parsedate: -> %s" % dt)
    return dt


# Create the db and fill it up with the values we need, taken out of
# the recoll records list
def recolltosql(conn, folders):
    rcldocs = folders.rcldocs()
    start = time.time()

    _createsqdb(conn)
    tabtorclfield = _prepareTags(conn)
    # uplog("Tagscreate: tabtorclfield: %s"%tabtorclfield)

    maxcnt = 0
    totcnt = 0
    c = conn.cursor()

    # A generic "Various Artists" tag value to be used for Albumartist if there are multiple artists
    # and no explicit AlbumArtist value. Set this as global, no need to query for it every time it's
    # needed.
    global variousartistsid
    variousartistsid = _auxtableinsert(conn, "artist", "Various Artists")

    for docidx in range(len(rcldocs)):
        doc = rcldocs[docidx]
        totcnt += 1

        if totcnt % 1000 == 0:
            time.sleep(0)

        # No need to include non-audio or non-tagged types
        if (
            doc["mtype"] not in audiomtypes
            or doc["mtype"] == "inode/directory"
            or doc["mtype"] == "audio/x-mpegurl"
        ):
            continue

        # Album creation ?
        album_id = _maybecreatealbum(conn, doc)

        trackno = _tracknofordoc(doc)

        path = uprclutils.docpath(doc)

        # Misc tag values:
        # Set base values for column names, values list, placeholders. Done this way because we used
        # to add to the lists for the tag values (which now use separate junction tables).
        columns = ["docidx", "album_id", "trackno", "title", "path"]
        values = [docidx, album_id, trackno, doc["title"], path]
        placehold = ["?", "?", "?", "?", "?"]
        for tb, rclfld in tabtorclfield:
            value = doc[rclfld]
            # Special processing for some fields
            if rclfld == "date":
                # See comment in parsedate
                if not value:
                    value = doc["dmtime"]
                if value:
                    value = parsedate(value)
            elif rclfld == "artist":
                value = doc["albumartist"]
                if not value:
                    value = doc["artist"]
            if not value:
                continue
            # rclaudio.py concatenates multiple values, using " | " as separator.
            valuelist = set(value.split(" | "))
            rowids = set()
            for value in valuelist:
                # Possibly insert in appropriate table (if value not already there), and
                # insert record in junction table for the corresponding field.
                rowid = _auxtableinsert(conn, tb, value)
                rowids.add(rowid)
                stmt = f"INSERT OR IGNORE INTO {_junctb(tb)}(docidx, {_clid(tb)}) VALUES (?, ?)"
                jvals = [docidx, rowid]
                # uplog(f"EXECUTING {stmt} values {jvals}")
                c.execute(stmt, jvals)
            if tb == "artist" and rowids:
                _updatealbartistlist(conn, album_id, rowids)
        # Create the main record in the tracks table.
        stmt = "INSERT INTO tracks(" + ",".join(columns) + ") VALUES(" + ",".join(placehold) + ")"
        c.execute(stmt, values)
    
    ## End Big doc loop

    #t1 = time.time()
    #uplog(f"recolltosql: docwalk: {t1-start:.1f} Seconds")
    _setalbumartists(conn)
    _setalbumcovers(conn, folders)
    #t2 = time.time()
    #uplog(f"recolltosql: setalbumxx: {t2-t1:.1f} Seconds")
    _createmergedalbums(conn)
    #t1 = time.time()
    #uplog(f"recolltosql: createmergedalbums: {t1-t2:.1f} Seconds")
    conn.commit()
    #t2 = time.time()
    #uplog(f"recolltosql: commit: {t2-t1:.1f} Seconds")
    rcldb = recoll.connect(confdir=uprclinit.getRclConfdir(), writable=True)
    _albumstorecoll(conn, rcldb)
    _artiststorecoll(conn, rcldb)
    end = time.time()
    uplog(f"recolltosql: processed {totcnt} docs in {end-start:.1f} Seconds")



# defined to emulate the session object from the tidal module, which is
# defined in the tidalapi part (we want to keep the qobuz api/ dir as
# much alone as possible.

import sys
import json
from datetime import datetime
from upmplgmodels import Artist, Album, Track, Playlist, SearchResult, Category, Genre
from upmplgutils import *
from qobuz.api import raw

# General limit for fetching stuff, in places where we do it in one chunk.
# The high limit set by Qobuz depends a bit on the nature of the request, but
# 100 works everywhere (some requests could use a bigger value).
general_slice = 100

# Show rate/bits with album titles? Simpler as a global
show_album_rate_and_bits = False


class Session(object):
    def __init__(self, format_id, fetch_resource_info=False, show_album_maxaudio=False):
        self.api = None
        self.user = None
        self.format_id = format_id
        self.fetch_resource_info = fetch_resource_info
        global show_album_rate_and_bits
        show_album_rate_and_bits = show_album_maxaudio

    def login(self, username, password, appid, cfvalue):
        # self.api will already exist if get_appid() was called first, which means that the
        # appid/secret were obtained through the spoofer, not the config. This happens when we are
        # working for OH Credentials. Else we create the API with our input params which come from
        # the configuration (and still may be empty in which case the spoofer will be used).
        if not self.api:
            self.api = raw.RawApi(appid, cfvalue)
        data = self.api.user_login(username=username, password=password)
        if data:
            self.user = User(self, self.api.user_id)
            return True
        else:
            return False

    def get_appid(self):
        if not self.api:
            self.api = raw.RawApi(None, None)
        return self.api.appid

    def get_appid_and_token(self):
        return (self.api.appid, self.api.user_auth_token)

    def get_media_url(self, trackid):
        info = self.get_media_info(trackid)
        return info["url"] if info and "url" in info else ""

    def get_media_info(self, trackid):
        info = self.api.track_getFileUrl(
            intent="stream", track_id=trackid, format_id=self.format_id
        )
        uplog(f"get_media_info got: {info}")
        return info

    def loopget(self, max, getter, getterka, parser):
        slice = 20
        offset = 0
        all = []
        while max == 0 or offset < max:
            getterka["limit"] = slice
            getterka["offset"] = offset
            jsdata = getter(**getterka)
            try:
                ndata = parser(jsdata)
                # uplog("loopget: %s" % json.dumps(ndata, indent=4))
            except Exception as err:
                uplog("loopget: exception while parsing: %s" % err)
                break
            all.extend(ndata)
            #uplog("GOT %d more (slice %d)" % (len(ndata), slice))
            if len(ndata) < slice:
                break
            offset += slice

        return all

    def get_album_tracks(self, albid):
        data = self.api.album_get(album_id=albid)
        album = _parse_album(data)
        return [_parse_track(self, t, albumarg=album) for t in data["tracks"]["items"]]

    def get_track(self, trackid):
        data = self.api.track_get(track_id=trackid)
        if data:
            return _parse_track(self, data)

    def get_playlist_tracks(self, plid):
        return self.loopget(
            1000,
            self.api.playlist_get,
            {"playlist_id": plid, "extra": "tracks"},
            lambda jsdata: [_parse_track(self, t) for t in jsdata["tracks"]["items"]],
        )

    def get_artist_albums(self, artid):
        data = self.api.artist_get(artist_id=artid, extra="albums")
        if "albums" in data:
            albums = [_parse_album(alb) for alb in data["albums"]["items"]]
            return [alb for alb in albums if alb.available]
        return []

    def get_artist_similar(self, artid):
        data = self.api.artist_getSimilarArtists(artist_id=artid)
        if "artists" in data and "items" in data["artists"]:
            return [_parse_artist(art) for art in data["artists"]["items"]]
        return []

    def get_featured_albums(self, genre_id="None", type="new-releases"):
        # uplog("get_featured_albums, genre_id %s type %s " % (genre_id, type))
        if genre_id != "None":
            data = self.api.album_getFeatured(
                type=type, genre_ids=genre_id + ":", limit=general_slice
            )
        else:
            data = self.api.album_getFeatured(type=type, limit=general_slice)

        try:
            albums = [_parse_album(alb) for alb in data["albums"]["items"]]
            if albums:
                return [alb for alb in albums if alb.available]
        except:
            pass
        return []

    def get_featured_playlists(self, genre_id="None", tags="None"):
        data = self.api.playlist_getFeatured(
            type="editor-picks", genre_ids=genre_id, tags=tags, limit=general_slice
        )
        if data and "playlists" in data:
            return [_parse_playlist(pl) for pl in data["playlists"]["items"]]
        return []

    def get_featured_artists(self):
        data = self.api.catalog_getFeatured(limit=general_slice)
        if "artists" in data:
            return [_parse_artist(i) for i in data["artists"]["items"]]
        return []

    def get_genres(self, parent=None):
        data = self.api.genre_list(parent_id=parent)
        return [Genre(id=None, name="All Genres")] + [
            _parse_genre(g) for g in data["genres"]["items"]
        ]

    # field : "artist", "album", "playlist", "title"
    # tp: 'tracks', 'albums', 'artists' or 'playlists'
    def _search1(self, field, query, tp):
        flt = None
        if tp == "tracks":
            if field == "artist":
                flt = _track_filter_artist
            elif field == "title":
                flt = _track_filter_title
            elif field == "album":
                flt = _track_filter_album
        elif tp == "albums":
            if field == "artist":
                flt = _album_filter_artist
            elif field == "album" or field == "title":
                flt = _album_filter_title
        elif tp == "artists":
            if field == "artist" or field == "title":
                flt = _artist_filter_name
        elif tp == "playlists":
            if field == "artist":
                return SearchResult(playlists=[])
            else:
                flt = _playlist_filter_name

        if tp == "artists":
            limit = 20
            slice = 20
        elif tp == "albums" or tp == "playlists":
            limit = 50
            slice = 50
        else:
            limit = 100
            slice = 100
            
        offset = 0
        all = []
        # The test on [(offset < x * limit) and ...] replaced a test on [offset < limit or...] to
        # work around the problem that if qobuz returns results, but on the wrong field, we
        # can be chasing our queue for very long
        while offset < (3 * limit) and len(all) < limit / 2:
            uplog(f"_search1: catsearch: tp {tp} q {query}, offset {offset} limit {slice}")
            data = self.api.catalog_search(query=query, type=tp, offset=offset, limit=slice)
            ncnt = 0
            ndata = []
            try:
                if tp == "artists":
                    ncnt = len(data["artists"]["items"])
                    ndata = [_parse_artist(i) for i in data["artists"]["items"]]
                    if flt:
                        ndata = [art for art in ndata if flt(art, query)]
                elif tp == "albums":
                    ncnt = len(data["albums"]["items"])
                    ndata = [_parse_album(i) for i in data["albums"]["items"]]
                    ndata = [alb for alb in ndata if alb.available]
                    if flt:
                        ndata = [alb for alb in ndata if flt(alb, query)]
                elif tp == "playlists":
                    # uplog("PLAYLISTS: %s" % json.dumps(data, indent=4))
                    ncnt = len(data["playlists"]["items"])
                    ndata = [_parse_playlist(i) for i in data["playlists"]["items"]]
                    if flt:
                        ndata = [pl for pl in ndata if flt(pl, query)]
                elif tp == "tracks":
                    ncnt = len(data["tracks"]["items"])
                    ndata = [_parse_track(self, i) for i in data["tracks"]["items"]]
                    if flt:
                        ndata = [track for track in ndata if flt(track, query)]
            except Exception as err:
                uplog("_search1: exception while parsing result: %s" % err)
                break
            all.extend(ndata)
            # uplog("Got %d more (slice %d)" % (ncnt, slice))
            if ncnt < slice:
                break
            offset += slice

        if tp == "artists":
            return SearchResult(artists=all)
        elif tp == "albums":
            return SearchResult(albums=all)
        elif tp == "playlists":
            return SearchResult(playlists=all)
        elif tp == "tracks":
            return SearchResult(tracks=all)

    def search(self, field, value, tp):
        uplog(f"search: field [{field}] value [{value}] tp [{tp}]")
        if tp:
            return self._search1(field, value, tp)
        else:
            cplt = SearchResult()
            if field != "album":
                res = self._search1(field, value, "artists")
                cplt.artists = res.artists
            res = self._search1(field, value, "albums")
            cplt.albums = res.albums
            res = self._search1(field, value, "tracks")
            cplt.tracks = res.tracks
            res = self._search1(field, value, "playlists")
            cplt.playlists = res.playlists
            return cplt


# The \b thing is useful because qobuz returns partial word matches, which I don't think is a good
# idea. Filter for full word matches instead.
def _search_in_field(fieldvalue, words):
    fieldvalue = strip_accents(fieldvalue)
    for w in strip_accents(words).split():
        if not re.search(r"\b" + w + r"\b", fieldvalue, re.IGNORECASE):
            return False
    return True

def _artist_filter_name(art, value):
    # uplog(f"_artist_filter_name: value [{value}] name [{art.name}]")
    return _search_in_field(art.name, value)

def _parse_artist(data):
    # uplog("_parse_artist: data %s" % data)
    kwargs = {"id": data["id"], "name": data["name"]}
    if "image" in data and data["image"] and "large" in data["image"]:
        kwargs["image"] = data["image"]["large"]
    return Artist(**kwargs)


def _parse_genre(data):
    return Genre(id=data["id"], name=data["name"])


def _album_filter_title(alb, value):
    #uplog(f"_album_filter_title: value [{value}] title [{alb.name}]")
    return _search_in_field(alb.name, value)

def _album_filter_artist(alb, value):
    #uplog(f"_album_filter_artist: value [{value}] title [{alb.name}]")
    return _search_in_field(alb.artist.name, value)


def _parse_album(json_obj, artist=None, artists=None):
    # uplog("Qobuz:_parse_album:DATA:\n%s"%json.dumps(json_obj, indent=4))
    if artist is None and "artist" in json_obj:
        artist = _parse_artist(json_obj["artist"])
    # if artists is None:
    #    artists = _parse_artists(json_obj['artists'])
    available = json_obj["streamable"] if "streamable" in json_obj else False
    # if not available:
    #    uplog("Album not streamable: %s " % json_obj['title'])
    name = json_obj["title"]
    if (
        show_album_rate_and_bits
        and "maximum_sampling_rate" in json_obj
        and "maximum_bit_depth" in json_obj
    ):
        name += (
            " ("
            + str(int(json_obj["maximum_sampling_rate"]))
            + "/"
            + str(json_obj["maximum_bit_depth"])
            + ")"
        )
    kwargs = {
        "id": json_obj["id"],
        "name": name,
        "num_tracks": json_obj.get("tracks_count"),
        "duration": json_obj.get("duration"),
        "artist": artist,
        "available": available,
        #'artists': artists,
    }
    if "image" in json_obj and "large" in json_obj["image"]:
        kwargs["image"] = json_obj["image"]["large"]

    # Note: we get no descriptions in the short albums from, e.g. search results, only when we
    # access /albums/get
    if "description" in json_obj:
        # uplog(f"_parse_album: setting description {json_obj['description'][0:80]}...")
        kwargs["description"] = json_obj["description"]

    try:
        if "release_date_original" in json_obj:
            # uplog(f"release_date_original: {json_obj['release_date_original']}")
            kwargs["release_date"] = json_obj["release_date_original"]
        elif "released_at" in json_obj:
            ts = json_obj["released_at"]
            dte = datetime.utcfromtimestamp(ts).strftime("%Y-%m-%d")
            # uplog(f"released_at: {ts} -> {dte}")
            kwargs["release_date"] = dte
        elif "releaseDate" in json_obj:
            #  I think that this is obsolete: not in qobuz data any more (if it ever was).
            kwargs["release_date"] = json_obj["releaseDate"]
    except Exception as ex:
        uplog(f"album: got exception {ex} while trying to set date")
        pass

    return Album(**kwargs)

def _playlist_filter_name(pl, value):
    #uplog(f"_playlist_filter_name: value [{value}] name [{pl.name}]")
    return _search_in_field(pl.name, value)

def _parse_playlist(data, artist=None, artists=None):
    # uplog("_parse_playlist: data %s" % data)
    kwargs = {
        "id": data["id"],
        "name": data["name"],
        "num_tracks": data.get("tracks_count"),
        "duration": data.get("duration"),
    }
    for suff in ("300", "150", ""):
        try:
            kwargs["image"] = data["images" + suff][0]
            break
        except:
            pass
    return Playlist(**kwargs)


def _track_filter_title(track, value):
    # uplog(f"_track_filter_title: value [{value}] title [{track.name}]")
    return _search_in_field(track.name, value)

def _track_filter_artist(track, value):
    return _search_in_field(track.artist.name, value)

def _track_filter_album(track, value):
    return _search_in_field(track.album.name, value)

def _parse_track(sess, json_obj, albumarg=None):
    artist = Artist()
    if "performer" in json_obj:
        artist = _parse_artist(json_obj["performer"])
    elif "artist" in json_obj:
        artist = _parse_artist(json_obj["artist"])
    elif albumarg and albumarg.artist:
        artist = albumarg.artist

    album = None
    if "album" in json_obj:
        album = _parse_album(json_obj["album"], artist)
    else:
        album = albumarg

    available = json_obj["streamable"] if "streamable" in json_obj else false
    # if not available:
    # uplog("Track not streamable: %s " % json_obj['title'])

    # artists = _parse_artists(json_obj['artists'])
    kwargs = {
        "id": json_obj["id"],
        "name": json_obj["title"],
        "duration": json_obj["duration"],
        "track_num": json_obj["track_number"],
        "disc_num": json_obj["media_number"],
        "artist": artist,
        "available": available,
        #'artists': artists,
    }
    if album:
        kwargs["album"] = album

    # Fetching the audio details is expensive (not held in the directory records, needs one call per
    # track), and few (if any?) renderers actually care about this information (they get it from the
    # stream when starting playing)
    if sess and sess.fetch_resource_info:
        trackinfo = sess.get_media_info(json_obj["id"])
        if "mime_type" in trackinfo:
            kwargs["mime"] = trackinfo["mime_type"]
            kwargs["samplefreq"] = trackinfo["sampling_rate"] * 1000
            kwargs["bitdepth"] = trackinfo["bit_depth"]
            kwargs["channels"] = 2  # Not supplied by get_media_info

    return Track(**kwargs)


class Favorites(object):

    def __init__(self, session):
        self.session = session

    def artists(self):
        return self.session.loopget(
            0,
            self.session.api.favorite_getUserFavorites,
            {"user_id": self.session.user.id, "type": "artists"},
            lambda r: [_parse_artist(item) for item in r["artists"]["items"]],
        )

    def albums(self):
        result = self.session.loopget(
            0,
            self.session.api.favorite_getUserFavorites,
            {"user_id": self.session.user.id, "type": "albums"},
            lambda r: [_parse_album(item) for item in r["albums"]["items"]],
        )
        return [alb for alb in result if alb.available]

    def playlists(self):
        return self.session.loopget(
            0,
            self.session.api.playlist_getUserPlaylists,
            {"user_id": self.session.user.id},
            lambda r: [_parse_playlist(item) for item in r["playlists"]["items"]],
        )

    def tracks(self):
        result = self.session.loopget(
            0,
            self.session.api.favorite_getUserFavorites,
            {"user_id": self.session.user.id, "type": "tracks"},
            lambda r: [_parse_track(self.session, item) for item in r["tracks"]["items"]],
        )
        return [trk for trk in result if trk.available]


class User(object):
    def __init__(self, session, id):
        self.session = session
        self.id = id
        self.favorites = Favorites(self.session)

    def playlists(self):
        return self.session.get_user_playlists(self.id)



#!/usr/bin/env python3

# Copyright (C) 2023,2024,2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import constants
from pathlib import Path


header: list[str] = [
    "# Configuration Parameters",
    "",
    "## Read this first",
    "",
    "The following is a non-exaustive list of the configuration options for the subsonic plugin.  ",
    "All the variable names must be prepended with `subsonic` when you add them to your upmpdcli.conf file.  ",
    "For boolean values (distinguishable because the default is True or False), you will have to use 1 or 0.",
    "",
    "## List of variables",
    "",
    "VARIABLE|DESCRIPTION|DEFAULT_VALUE",
    ":---|:---|:---"
    ]


def generate():
    cnt: int = 0
    print("Generating ...")
    script_dir = Path(__file__).parent.absolute()
    file_path = script_dir / "config.md"
    with open(file_path, "w", encoding="utf-8") as f:
        curr: str
        for curr in header:
            f.write(f"{curr}\n")
    with open(file_path, "a", encoding="utf-8") as f:
        config_param: constants.ConfigParam
        for config_param in constants.ConfigParam:
            f.write(f"{config_param.key}|{config_param.description}|{config_param.default_value}\n")
            cnt += 1
    print(f"Generation complete [{cnt} entries])")


if __name__ == "__main__":
    generate()



# Copyright (C) 2016-2021 J.F.Dockes
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the
#   Free Software Foundation, Inc.,
#   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
#

"""Shared code for the plugins which reuse the routing module from the Tidal Kodi plugin."""

import re
import posixpath

from upmplgutils import uplog
import upmplgutils


# Semi-bogus class instanciated as global object for helping with reusing kodi addon code
class XbmcPlugin:
    def __init__(self, idprefix, pid, offset=0, count=0, routeplugin=None):
        self.idprefix = idprefix
        upmplgutils.setidprefix(idprefix)
        self.routeplug = routeplugin
        self.objid = pid
        self.offset = offset
        self.count = count

        self.entries = []
        self.total = 0

    # item_num is the number to be prepended to the title, if different from None
    # this is for the benefit of kodi which always sort entries by name
    def add_directory(self, title, endpoint, arturi=None, item_num=None):
        if callable(endpoint):
            endpoint = self.routeplug.url_for(endpoint)
        if item_num:
            title = f"[{item_num:02}] {title}"
        e = upmplgutils.direntry(
            self.idprefix + endpoint, self.objid, title, arturi=arturi
        )
        self.entries.append(e)

    def urls_from_id(self, view_func, items):
        if not items:
            return []
        uplog("urls_from_id: items: %s" % str([item.id for item in items]), level=5)
        return [
            self.routeplug.url_for(view_func, item.id)
            for item in items
            if str(item.id).find("http") != 0
        ]

    # initial_item_num is the start number to be prepended to the title, if
    # different from None
    # This is for the benefit of kodi which always sort entries by name
    def view(self, data_items, urls, end=True, initial_item_num=None):
        if not data_items or not urls:
            return
        for item, url in zip(data_items, urls):
            title = item.name
            try:
                image = item.image if item.image else None
            except:
                image = None
            try:
                upnpclass = item.upnpclass if item.upnpclass else None
            except:
                upnpclass = None
            try:
                artnm = item.artist.name if item.artist.name else None
            except:
                artnm = None
            try:
                description = item.description if item.description else None
            except:
                description = None

            if initial_item_num:
                title = f"[{initial_item_num:02}] {title}"
                initial_item_num += 1
            self.entries.append(
                upmplgutils.direntry(
                    self.idprefix + url,
                    self.objid,
                    title,
                    arturi=image,
                    artist=artnm,
                    upnpclass=upnpclass,
                    description=description,
                )
            )



# Copyright (C) 2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class UpMpdMeta(Enum):
    ALBUM_QUALITY = "albumquality"
    ALBUM_VERSION = "albumversion"
    ALBUM_EXPLICIT_STATUS = "albumexplicitstatus"
    ALBUM_ID = "albumid"
    ALBUM_MUSICBRAINZ_ID = "albummusicbrainzid"
    ALBUM_RECORD_LABELS = "albumrecordlabels"
    ALBUM_DURATION = "albumduration"
    ALBUM_DISC_AND_TRACK_COUNTERS = "albumdisctrackcounters"
    ALBUM_ARTIST = "albumartist"
    # ALBUM_ARTISTS = "albumartists"
    ALBUM_TITLE = "albumtitle"
    ALBUM_YEAR = "albumyear"
    ALBUM_MEDIA_TYPE = "albummediatype"
    MOOD = "mood"
    ALBUM_RELEASE_DATE = "albumreleasedate"
    ALBUM_ORIGINAL_RELEASE_DATE = "albumoriginalreleasedate"
    ALBUM_AVAILABLE_RELEASE_DATE = "albumavailablereleasedate"
    IS_COMPILATION = "albumiscompilation"
    RELEASE_TYPES = "albumreleasetypes"
    ARTIST_ID = "artistid"
    ARTIST_MUSICBRAINZ_ID = "artistmusicbrainzid"
    ARTIST_ALBUM_COUNT = "artistalbumcount"
    ARTIST_MEDIA_TYPE = "artistmediatype"
    ARTIST_ROLE = "artistrole"
    ALBUM_PATH = "albumpath"
    TRACK_DURATION = "trackduration"
    TRACK_NUMBER = "tracknumber"
    TRACK_LAST_PLAYED = "tracklastplayed"
    DISC_NUMBER = "discnumber"
    COPYRIGHT = "copyright"
    UNIVERSAL_PRODUCT_NUMBER = "universalproductnumber"
    TRACK_QUALITY = "trackquality"
    ALBUM_QUALITY_OF_TRACKS = "albumqualityoftracks"
    ALBUM_LAST_PLAYED = "albumlastplayed"
    ALBUM_COMPRESSION_TYPE = "albumcompressiontype"
    ALBUM_AVERAGE_BITRATE = "albumaveragebitrate"


def get_duration_display_from_sec(duration_sec: int) -> str:
    if duration_sec < 0:
        # duration is invalid
        return "<invalid duration>"
    # hours, minutes, seconds
    remaining_sec: int = duration_sec
    seconds: int = duration_sec % 60
    remaining_sec -= seconds
    minutes: int = int(int(remaining_sec / 60) % 60)
    remaining_sec -= (minutes * 60)
    hours: int = int(remaining_sec / 3600)
    result: str = ""
    # format it!
    if hours > 0:
        result += f"{hours}h"
    if minutes > 0:
        if len(result) > 0:
            result += " "
        result += f"{minutes:02d}m"
    # add seconds in any case
    if len(result) > 0:
        result += " "
    result += f"{seconds:02d}s"
    return result



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import constants

import upmplgutils
# uncomment when py-sonic with useragent support is released
from subsonic_connector.connector import Constants as SubsonicConnectorConstants
from subsonic_connector.configuration import ConfigurationInterface
from tag_type import TagType
from dotenv import dotenv_values


def get_plugin_config_variable_name(name: str) -> str:
    return f"{constants.PluginConstant.PLUGIN_NAME.value}{name}"


def _get_option_value(nm, dflt: any = None):
    return upmplgutils.getOptionValue(get_plugin_config_variable_name(nm), dflt)


def _get_option_value_as_bool(nm: str, default_value: int) -> bool:
    return _get_option_value(nm, default_value) == 1


def get_whitelist_codecs() -> list[str]:
    return str(get_config_param_as_str(constants.ConfigParam.WHITELIST_CODECS)).split(",")


# supported unless initializer understands it is not
# begin
album_list_by_highest_supported: bool = True
internet_radio_stations_supported: bool = True
# end


def is_transcode_enabled() -> bool:
    return (get_config_param_as_str(constants.ConfigParam.TRANSCODE_CODEC) or
            get_config_param_as_str(constants.ConfigParam.TRANSCODE_MAX_BITRATE))


def get_transcode_codec() -> str:
    transcode_codec: str = get_config_param_as_str(constants.ConfigParam.TRANSCODE_CODEC)
    if transcode_codec:
        return transcode_codec
    # only bitrate might be specified, in this case we return a fallback value for transcode codec
    if is_transcode_enabled():
        return constants.Defaults.FALLBACK_TRANSCODE_CODEC.value
    return None


def get_transcode_max_bitrate() -> int:
    transcode_max_bitrate: str = get_config_param_as_str(constants.ConfigParam.TRANSCODE_MAX_BITRATE)
    if transcode_max_bitrate:
        return int(transcode_max_bitrate)
    # if transcode codec is specified, we return the default, if available
    transcode_codec: str = get_config_param_as_str(constants.ConfigParam.TRANSCODE_CODEC)
    if transcode_codec:
        return constants.get_default_bitrate_by_codec(transcode_codec)
    # nothing to do
    return None


def is_tag_supported(tag: TagType) -> bool:
    # true unless there are exceptions ...
    if tag == TagType.HIGHEST_RATED_ALBUMS:
        return album_list_by_highest_supported
    elif tag == TagType.INTERNET_RADIOS:
        return internet_radio_stations_supported
    return True


def get_config_param_as_str(configuration_parameter: constants.ConfigParam) -> str:
    dv: str | None = configuration_parameter.default_value
    if dv is not None and not isinstance(dv, str):
        raise Exception(f"Invalid default value for [{configuration_parameter.key}]")
    v: any = _get_option_value(configuration_parameter.key, dv)
    if v is None:
        return None
    # v is set, check type!
    if isinstance(v, str):
        return v
    # try to convert to string
    return str(v)


def get_config_param_as_int(configuration_parameter: constants.ConfigParam) -> str:
    dv: int | None = configuration_parameter.default_value
    if dv is not None and not isinstance(dv, int):
        raise Exception(f"Invalid default value for [{configuration_parameter.key}]")
    v: any = _get_option_value(configuration_parameter.key, dv)
    if v is None:
        return None
    # v is set, check type!
    if isinstance(v, int):
        return v
    # try to convert to int
    return int(v)


def get_config_param_as_bool(configuration_parameter: constants.ConfigParam) -> bool:
    default_value_as_int: int = 0
    dv: any = configuration_parameter.default_value
    if isinstance(dv, int):
        default_value_as_int = 1 if dv == 1 else 0
    elif isinstance(dv, bool):
        default_value_as_int = 1 if dv else 0
    return _get_option_value_as_bool(
        configuration_parameter.key,
        default_value_as_int)


def get_items_per_page() -> int:
    return min(
        constants.Defaults.SUBSONIC_API_MAX_RETURN_SIZE.value,
        get_config_param_as_int(constants.ConfigParam.ITEMS_PER_PAGE))


def getWebServerDocumentRoot() -> str:
    return upmplgutils.getUpnpWebDocRoot(constants.PluginConstant.PLUGIN_NAME.value)


def get_webserver_path_images() -> list[str]:
    return [
        constants.PluginConstant.PLUGIN_NAME.value,
        "images"]


def get_webserver_path_images_static() -> list[str]:
    return get_webserver_path_images() + ["static"]


def get_webserver_path_images_cache() -> list[str]:
    return get_webserver_path_images() + ["cache"]


def get_custom_headers() -> dict[str, str]:
    custom_headers: dict[str, str] = {}
    custom_headers_file_name: str = get_config_param_as_str(constants.ConfigParam.CUSTOM_HEADERS_FILE_NAME)
    # msgproc.log(f"UpmpdcliSubsonicConfig custom_headers_file_name [{custom_headers_file_name}]")
    if custom_headers_file_name:
        loaded_custom_headers: dict[str, str] = dotenv_values(custom_headers_file_name)
        custom_headers.update(loaded_custom_headers)
    return custom_headers


class UpmpdcliSubsonicConfig(ConfigurationInterface):

    def __init__(self):
        self.__custom_headers: dict[str, str] = get_custom_headers()

    def getBaseUrl(self) -> str: return _get_option_value("baseurl")

    def getPort(self) -> int: return _get_option_value("port")

    def getServerPath(self) -> int: return _get_option_value("serverpath")

    def getUserName(self) -> str: return _get_option_value("user")

    def getPassword(self) -> str: return _get_option_value("password")

    def getLegacyAuth(self) -> bool:
        legacy_auth_enabled_str: str = _get_option_value("legacyauth", "false")
        if not legacy_auth_enabled_str.lower() in ["true", "false", "1", "0"]:
            raise Exception(f"Invalid value for SUBSONIC_LEGACYAUTH [{legacy_auth_enabled_str}]")
        return legacy_auth_enabled_str in ["true", "1"]

    def getSalt(self) -> str:
        return None

    def getToken(self) -> str:
        return None

    def getUserAgent(self) -> bool:
        # only use empty user agent if explicitly instructed to do so
        if get_config_param_as_bool(constants.ConfigParam.SKIP_USER_AGENT):
            return None
        user_agent_str: str = get_config_param_as_str(constants.ConfigParam.USER_AGENT)
        if user_agent_str is None or len(user_agent_str) == 0:
            user_agent_str = constants.ConfigParam.USER_AGENT.default_value
        return user_agent_str

    def getCustomHeaders(self) -> dict[str, str]:
        return self.__custom_headers

    def getApiVersion(self) -> str: return SubsonicConnectorConstants.API_VERSION.value

    def getAppName(self) -> str: return "upmpdcli-subsonic-plugin"



# Copyright (C) 2023 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import upmplgmodels


def set_track_number(track_number : str, target : dict):
    target['upnp:originalTrackNumber'] = track_number


def set_album_art_from_uri(album_art_uri : str, target : dict):
    target['upnp:albumArtURI'] = album_art_uri


def set_album_title(album_title : str, target : dict):
    target['tt'] = album_title


def set_album_id(album_id : str, target : dict):
    target['album_id'] = album_id


def set_class(upnp_class : str, target : dict):
    target['upnp:class'] = upnp_class


def set_artist(artist : str, target : dict):
    target['upnp:artist'] = artist


def set_class_music_track(target : dict):
    target['upnp:class'] = upmplgmodels.Track.upnpclass


def set_class_album(target : dict):
    target['upnp:class'] = upmplgmodels.Album.upnpclass


def set_class_artist(target : dict):
    target['upnp:class'] = upmplgmodels.Artist.upnpclass


def set_bit_depth(bit_depth: int, target: dict):
    target['res:bitsPerSample'] = str(bit_depth)


def set_sample_rate(sample_rate: int, target: dict):
    target['res:samplefreq'] = str(sample_rate)


def set_channel_count(channel_count: int, target: dict):
    target['res:channels'] = str(channel_count)


def calc_bitrate(num_channels: int, bit_depth: int, sampling_rate: int) -> int:
    return int((num_channels * bit_depth * sampling_rate) / 1000)


def set_bit_rate(bit_rate: int, target: dict):
    target['kbs'] = str(bit_rate)



#!/usr/bin/env python3

# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import requests
import json
import sys
import os

apiurl = "https://motherearth.streamserver24.com/api/nowplaying"

def msg(s):
    print(f"{s}", file=sys.stderr)
def usage():
    msg(f"Usage: {os.path.basename(sys.argv[0])} <station_shortcode>")
    msg(" Station shortcodes: motherearth_klassik, motherearth_instrumental, motherearth_jazz, motherearth\n")
    sys.exit(1)

if len(sys.argv) < 2:
    usage()
shortcode = sys.argv[1]

try:
    r = requests.get(apiurl)
    r.raise_for_status()
    jsd = r.json()
except:
    sys.exit(1)
      
for station in jsd:
    try:
        sc = station["station"]["shortcode"]
        #msg(sc)
        if sc == shortcode:
            nowplaying = station["now_playing"]
            data = {}
            data["title"] = nowplaying["song"]["title"]
            data["artist"] = nowplaying["song"]["artist"]
            data["artUrl"] = nowplaying["song"]["art"]
            data["reload"] = int(nowplaying["remaining"])
            print(f"{json.dumps(data)}")
    except:
        sys.exit(1)
